# SPDX-License-Identifer: GPL-3.0-or-later

import logging
import os
import sys
from pathlib import PurePath
from typing import Any


class LoggerFactory:
    DEFAULT_LOGLEVEL = getattr(
        logging,
        os.getenv("APT_MIRROR_LOGLEVEL", "info").upper(),
    )
    DEFAULT_FORMAT = (
        "%(asctime)s: [%(process)d] %(levelname)s %(name_abbr)s %(message)s"
    )
    FILES: dict[Any, PurePath] = {}
    FILE_HANDLERS: dict[Any, logging.FileHandler] = {}

    @staticmethod
    def init_logging():
        logging.basicConfig(
            format=LoggerFactory.DEFAULT_FORMAT,
            level=LoggerFactory.DEFAULT_LOGLEVEL,
            stream=sys.stderr,
        )
        logging.getLogger().handlers[0].addFilter(NameAbbrFilter())

        if LoggerFactory.DEFAULT_LOGLEVEL != logging.DEBUG:
            logging.getLogger("httpx").setLevel(logging.WARNING)

        logging.debug("Logging started")

    @staticmethod
    def get_logger(
        obj: Any,
        logger_id: Any | None = None,
    ) -> logging.Logger:
        log_name = (
            ".".join((obj.__class__.__module__, obj.__class__.__qualname__))
            if not isinstance(obj, str)
            else obj
        )

        log = logging.getLogger(
            f"{log_name}.{hash(logger_id)}" if logger_id else log_name
        )
        log.name = log_name
        log.level = LoggerFactory.DEFAULT_LOGLEVEL

        if logger_id in LoggerFactory.FILES:
            if logger_id not in LoggerFactory.FILE_HANDLERS:
                file_handler = logging.FileHandler(
                    LoggerFactory.FILES[logger_id],
                    mode="w",
                    encoding="utf-8",
                )
                file_handler.setFormatter(
                    logging.Formatter(LoggerFactory.DEFAULT_FORMAT)
                )
                file_handler.addFilter(NameAbbrFilter())

                LoggerFactory.FILE_HANDLERS[logger_id] = file_handler

            log.addHandler(LoggerFactory.FILE_HANDLERS[logger_id])

        return log

    @staticmethod
    def add_log_file(logger_id: Any, file: PurePath):
        LoggerFactory.FILES[logger_id] = file


class NameAbbrFilter(logging.Filter):
    def filter(self, record: logging.LogRecord):
        modules = record.name.split(".")
        record.name_abbr = ".".join(
            ["_".join(p[:1] for p in m.split("_")) for m in modules[:-1]]
            + [modules[-1]]
        )

        return True


LoggerFactory.init_logging()



# SPDX-License-Identifer: GPL-3.0-or-later


from .downloader import Downloader, DownloaderSettings
from .protocols.ftp import FTPDownloader
from .protocols.http import HTTPDownloader


class UnsupportedURLException(ValueError):
    pass


class DownloaderFactory:
    @staticmethod
    def for_settings(*, settings: DownloaderSettings) -> Downloader:
        if settings.url.scheme.startswith("http"):
            cls = HTTPDownloader
        elif settings.url.scheme.startswith("ftp"):
            cls = FTPDownloader
        else:
            raise UnsupportedURLException(
                f"Unsupported URL scheme: {settings.url.scheme}"
            )

        return cls(settings=settings)



# SPDX-License-Identifer: GPL-3.0-or-later

from contextlib import asynccontextmanager
from datetime import datetime
from email.utils import parsedate_to_datetime
from pathlib import Path

import httpx

from ..downloader import Downloader
from ..response import DownloadResponse


class HTTPDownloader(Downloader):
    def __post_init__(self):
        auth = None
        if self._settings.url.username and self._settings.url.password:
            auth = (self._settings.url.username, self._settings.url.password)

        base_url = str(self._settings.url)
        if not base_url.endswith("/"):
            base_url += "/"

        http_limits = httpx.Limits(
            max_connections=256,
            max_keepalive_connections=32,
            keepalive_expiry=5,
        )

        proxy_mounts: dict[str, httpx.AsyncHTTPTransport] = {}
        for scheme in ("http://", "https://"):
            proxy = self._settings.proxy.for_scheme(scheme)
            proxy_mounts[scheme] = httpx.AsyncHTTPTransport(
                verify=self._settings.verify_ca_certificate,
                http1=True,
                http2=not self._settings.http2_disable,
                limits=http_limits,
                proxy=httpx.Proxy(proxy) if proxy else None,
                retries=5,
            )

        client_certificate = None
        if self._settings.client_certificate:
            if self._settings.client_private_key:
                client_certificate = (
                    self._settings.client_certificate,
                    self._settings.client_private_key,
                )
            else:
                client_certificate = self._settings.client_certificate

        self._httpx = httpx.AsyncClient(
            base_url=base_url,
            auth=auth,
            timeout=httpx.Timeout(
                15,
                connect=30,
                read=60,
            ),
            follow_redirects=True,
            mounts=proxy_mounts,
            transport=httpx.AsyncHTTPTransport(
                verify=self._settings.verify_ca_certificate,
                cert=client_certificate,
                http1=True,
                http2=not self._settings.http2_disable,
                limits=http_limits,
                retries=5,
            ),
            max_redirects=5,
            headers={
                "Accept-Encoding": "identity",
                "Cache-Control": "no-cache",
                "Pragma": "no-cache",
                "User-Agent": self._settings.user_agent,
            },
        )

    def aiter_bytes(self, response: httpx.Response):
        def func():
            return response.aiter_bytes(chunk_size=self.BUFFER_SIZE)

        return func

    @asynccontextmanager
    async def stream(self, source_path: Path):
        try:
            async with (
                self._httpx.stream("GET", str(source_path)) as response,
            ):
                date: datetime | None
                try:
                    date = parsedate_to_datetime(  # type: ignore
                        response.headers.get("Last-Modified")
                    )
                except ValueError:
                    date = None

                try:
                    size = int(response.headers.get("Content-Length"))
                except (TypeError, ValueError):
                    size = None

                yield DownloadResponse(
                    missing=response.is_client_error,
                    error=(
                        f"HTTP/{response.status_code}"
                        if response.is_server_error
                        else None
                    ),
                    date=date,  # type: ignore
                    size=size,
                    _stream=self.aiter_bytes(response),
                )
        except httpx.RemoteProtocolError as ex:
            # https://github.com/encode/httpx/discussions/2056
            server_disconnected = (
                bool(ex.args) and "Server disconnected" not in ex.args[0]
            )

            yield DownloadResponse(
                _stream=None,
                retry=server_disconnected,
                error=str(ex) if not server_disconnected else None,
            )
        except Exception as ex:  # pylint: disable=W0718
            yield DownloadResponse(
                _stream=None,
                error=f"{ex.__class__.__qualname__}: {str(ex)}",
            )



# SPDX-License-Identifer: GPL-3.0-or-later

from .download_file import (
    DownloadFile,
    DownloadFileCompressionVariant,
    FileCompression,
    HashSum,
    HashType,
)
from .downloader import Downloader
from .factory import DownloaderFactory
from .proxy import Proxy
from .url import URL

__all__ = [
    "Downloader",
    "DownloaderFactory",
    "DownloadFile",
    "DownloadFileCompressionVariant",
    "FileCompression",
    "HashSum",
    "HashType",
    "Proxy",
    "URL",
]



import platform
from unittest import IsolatedAsyncioTestCase

from caio import linux_aio_asyncio, thread_aio_asyncio


class TestAIOFile(IsolatedAsyncioTestCase):
    def __init__(self, methodName: str = "runTest") -> None:
        super().__init__(methodName)
        self._platform = platform.system().lower()

    def linux_aio_supported(self):
        if "linux" not in self._platform:
            return False

        if platform.python_implementation() != "CPython":
            return False

        return True

    def thread_aio_supported(self):
        if "linux" not in self._platform and "darwin" not in self._platform:
            return False

        if platform.python_implementation() != "CPython":
            return False

    async def test_linux_aio(self):
        if not self.linux_aio_supported():
            return

        async with linux_aio_asyncio.AsyncioContext() as _:
            pass

    async def test_thread_aio(self):
        if not self.thread_aio_supported():
            return

        async with thread_aio_asyncio.AsyncioContext() as _:
            pass



# SPDX-License-Identifer: GPL-3.0-or-later

from dataclasses import dataclass
from pathlib import Path
from urllib import parse


@dataclass
class URL:
    scheme: str
    netloc: str
    path: str
    params: str
    query: str
    fragment: str
    hostname: str | None
    port: int | None
    username: str | None
    password: str | None

    @classmethod
    def from_string(cls, url_string: str):
        url = parse.urlparse(url_string)
        return cls(
            scheme=url.scheme,
            netloc=url.netloc,
            path=url.path,
            params=url.params,
            query=url.query,
            fragment=url.fragment,
            hostname=url.hostname,
            port=url.port,
            username=url.username,
            password=url.password,
        )

    def get_host(self):
        _, _, host = self.netloc.rpartition("@")
        return host

    def without_auth(self):
        return parse.urlunparse(
            (
                self.scheme,
                self.get_host(),
                self.path,
                self.params,
                self.query,
                self.fragment,
            )
        )

    def as_filesystem_path(self, encode_tilde: bool) -> Path:
        path = Path(self.get_host()) / self.path.lstrip("/")

        if encode_tilde:
            return Path(str(path).replace("~", "%7E"))

        return path

    def for_path(self, path: Path | str) -> str:
        str_path = str(path)

        base_path = self.path
        if base_path.endswith("/"):
            base_path = base_path[:-1]

        if str_path.startswith("/"):
            str_path = str_path[1:]

        return parse.urlunparse(
            (
                self.scheme,
                self.get_host(),
                f"{base_path}/{str_path}",
                self.params,
                self.query,
                self.fragment,
            )
        )

    def is_part_of(self, url: "URL"):
        return str(url).startswith(str(self))

    def __str__(self) -> str:
        return self.without_auth()

    def __hash__(self) -> int:
        return hash(self.without_auth())

    def __eq__(self, __value: object) -> bool:
        if not isinstance(__value, URL):
            return False

        return self.without_auth() == __value.without_auth()



# SPDX-License-Identifer: GPL-3.0-or-later


class PackageFilter:
    def __init__(self) -> None:
        self.include_source_name: set[str] = set()
        self.exclude_source_name: set[str] = set()
        self.include_binary_packages: set[str] = set()
        self.exclude_binary_packages: set[str] = set()

    def package_allowed(
        self, source_name: str, package_name: str | None = None
    ) -> bool:
        if self.include_source_name and source_name not in self.include_source_name:
            return False

        if self.exclude_source_name and source_name in self.exclude_source_name:
            return False

        if package_name:
            if (
                self.include_binary_packages
                and package_name not in self.include_binary_packages
            ):
                return False

            if (
                self.exclude_binary_packages
                and package_name in self.exclude_binary_packages
            ):
                return False

        return True



# SPDX-License-Identifer: GPL-3.0-or-later

from dataclasses import dataclass
from urllib import parse


@dataclass
class Proxy:
    use_proxy: bool
    http_proxy: str | None
    https_proxy: str | None
    username: str | None
    password: str | None

    def for_scheme(self, scheme: str) -> str | None:
        if not self.use_proxy:
            return None

        if scheme == "http://" and self.http_proxy:
            return self.url_for_proxy(self.http_proxy)

        if scheme == "https://" and self.https_proxy:
            return self.url_for_proxy(self.https_proxy)

        return None

    def url_for_proxy(self, proxy: str) -> str:
        if "://" not in proxy:
            proxy = f"http://{proxy}"

        url = parse.urlparse(proxy)
        if self.username:
            auth = parse.quote(self.username, safe="")
            if self.password:
                auth = f"{auth}:{parse.quote(self.password, safe='')}"

            url = url._replace(netloc=f"{auth}@{url.netloc}")

        return parse.urlunparse(url)



from pathlib import Path

from apt_mirror.apt_mirror import PathCleaner
from apt_mirror.config import Config, RepositoryConfigException
from apt_mirror.download.url import URL
from apt_mirror.netrc import NetRC
from apt_mirror.repository import ByHash
from tests.base import BaseTest


class TestConfig(BaseTest):
    def test_multiple_codenames(self):
        config = self.get_config("MixedConfig")

        self.assertEqual(len(config.repositories), 4)

        debian_security = self.ensure_repository(
            config.repositories["http://ftp.debian.org/debian-security"]
        )

        self.assertCountEqual(
            debian_security.codenames,
            ("bookworm-security", "bullseye-security", "trixie-security"),
        )

        self.assertFalse(
            debian_security.codenames["bookworm-security"].should_mirror_source()
        )
        self.assertFalse(
            debian_security.codenames["trixie-security"].should_mirror_source()
        )
        self.assertTrue(
            debian_security.codenames["bullseye-security"].should_mirror_source()
        )

        self.assertCountEqual(
            debian_security.codenames["trixie-security"].components["main"].arches,
            {
                config.default_arch,
                "amd64",
                "arm64",
                "armel",
                "armhf",
                "i386",
                "mips64el",
                "ppc64el",
                "s390x",
            },
        )

        self.assertCountEqual(
            debian_security.codenames["bullseye-security"].components["main"].arches,
            {
                config.default_arch,
                "amd64",
                "arm64",
                "armel",
                "armhf",
                "i386",
                "mips64el",
                "ppc64el",
                "s390x",
            },
        )

        self.assertCountEqual(
            debian_security.codenames["bookworm-security"].components["main"].arches,
            {config.default_arch, "arm64"},
        )

        ubuntu_security = self.ensure_repository(
            config.repositories["http://archive.ubuntu.com/ubuntu"]
        )

        self.assertCountEqual(
            ubuntu_security.codenames,
            (
                "mantic",
                "mantic-security",
                "mantic-updates",
                "mantic-backports",
                "noble",
                "noble-security",
                "noble-updates",
                "noble-backports",
                "oracular",
                "oracular-security",
                "oracular-updates",
                "oracular-backports",
            ),
        )

        flat_repository = self.ensure_flat_repository(
            config.repositories["http://mirror.something.ru/repository"]
        )

        self.assertTrue(flat_repository.is_binaries_enabled)
        self.assertEqual(
            str(flat_repository.url), "http://mirror.something.ru/repository"
        )
        self.assertCountEqual(flat_repository.directories.keys(), [Path("subpath")])
        self.assertTrue(
            flat_repository.directories[Path("subpath")].mirror_binaries, True
        )

    def test_codenames_list(self):
        config = self.get_config("MixedConfig")

        proxmox_apqa = self.ensure_repository(
            config.repositories["https://mirrors.apqa.cn/proxmox/debian/pve"]
        )

        self.assertCountEqual(
            proxmox_apqa.codenames,
            ("bookworm", "bullseye"),
        )

        self.assertCountEqual(
            proxmox_apqa.codenames["bookworm"].components["port"].arches,
            {
                config.default_arch,
                "amd64",
                "arm64",
                "i386",
            },
        )

        self.assertCountEqual(
            proxmox_apqa.codenames["bullseye"].components["port"].arches,
            {
                config.default_arch,
                "amd64",
                "arm64",
                "i386",
            },
        )

    def test_skip_clean(self):
        config = self.get_config("SkipCleanConfig")

        debian = self.ensure_repository(
            config.repositories["http://ftp.debian.org/debian"]
        )

        self.assertCountEqual(debian.skip_clean, (Path("abcd"), Path("def/def")))

        cleaner = PathCleaner(
            self.TEST_DATA / "SkipCleanConfig" / "dir1", debian.skip_clean
        )
        self.assertEqual(cleaner.clean_files_count, 0)
        self.assertEqual(cleaner.folders_count, 0)

        cleaner = PathCleaner(
            self.TEST_DATA / "SkipCleanConfig" / "dir2", debian.skip_clean
        )
        self.assertEqual(cleaner.clean_files_count, 1)
        self.assertEqual(cleaner.folders_count, 1)

    def test_http2_disable(self):
        config = self.get_config("MixedConfig")

        debian = self.ensure_repository(
            config.repositories["http://ftp.debian.org/debian-security"]
        )

        self.assertTrue(debian.http2_disable)

        ubuntu = self.ensure_repository(
            config.repositories["http://archive.ubuntu.com/ubuntu"]
        )

        self.assertFalse(ubuntu.http2_disable)

    def test_by_hash(self):
        config = self.get_config("ByHashConfig")

        repository = self.ensure_repository(
            config.repositories["http://ftp.debian.org/debian-security"]
        )

        self.assertEqual(repository.codenames["trixie-security"].by_hash, ByHash.FORCE)
        self.assertEqual(repository.codenames["bookworm-security"].by_hash, ByHash.NO)
        self.assertEqual(repository.codenames["stretch-security"].by_hash, ByHash.YES)

        repository = self.ensure_repository(
            config.repositories["http://archive.ubuntu.com/ubuntu"]
        )

        self.assertEqual(repository.codenames["mantic"].by_hash, ByHash.NO)

        repository = self.ensure_repository(
            config.repositories["http://mirror.something.ru/repository"]
        )

        self.assertEqual(repository.codenames["codename"].by_hash, ByHash.NO)

    def test_dist_upgrader(self):
        config = self.get_config("MixedConfig")

        repository = self.ensure_flat_repository(
            config.repositories["http://mirror.something.ru/repository"]
        )

        self.assertFalse(repository.mirror_dist_upgrader)

        repository = self.ensure_repository(
            config.repositories["http://archive.ubuntu.com/ubuntu"]
        )

        self.assertTrue(repository.mirror_dist_upgrader)

    def test_clean(self):
        config = self.get_config("MixedConfig")

        repository = self.ensure_flat_repository(
            config.repositories["http://mirror.something.ru/repository"]
        )

        self.assertTrue(repository.clean, True)

    def test_src_arch(self):
        config = self.get_config("SrcOptionConfig")

        repository = self.ensure_repository(
            config.repositories["http://ftp.debian.org/debian-security"]
        )

        self.assertFalse(repository.is_binaries_enabled)
        self.assertTrue(repository.is_source_enabled)

        repository = self.ensure_repository(
            config.repositories["http://archive.ubuntu.com/ubuntu"]
        )

        self.assertFalse(repository.is_binaries_enabled)
        self.assertTrue(repository.is_source_enabled)

    def test_ignore_errors(self):
        config = self.get_config("IgnoreErrorsConfig")

        repository = self.ensure_repository(
            config.repositories[
                "https://packages.gitlab.com/runner/gitlab-runner/debian"
            ]
        )

        self.assertCountEqual(
            repository.ignore_errors,
            [
                "pool/bullseye/main/g/gitlab-runner/gitlab-runner_14.8.1_amd64.deb",
                "pool/bullseye/main/d",
            ],
        )

    def test_multiple_flat_directories(self):
        config = self.get_config("FlatConfig")

        repository = self.ensure_flat_repository(
            config.repositories["https://packages.ntop.org/apt-stable/20.04/"]
        )

        self.assertEqual(len(repository.directories), 3)

        self.assertTrue(repository.directories[Path("x64")].mirror_binaries)
        self.assertFalse(repository.directories[Path("x64")].mirror_source)

        self.assertFalse(repository.directories[Path("x32")].mirror_binaries)
        self.assertTrue(repository.directories[Path("x32")].mirror_source)

        self.assertTrue(repository.directories[Path("all")].mirror_binaries)
        self.assertFalse(repository.directories[Path("all")].mirror_source)

        repository = self.ensure_flat_repository(
            config.repositories["http://localhost:8080/repo"]
        )

        self.assertEqual(len(repository.directories), 2)

        self.assertTrue(repository.directories[Path("bin1/")].mirror_binaries)
        self.assertTrue(repository.directories[Path("bin1/")].mirror_source)

        self.assertTrue(repository.directories[Path("bin2/")].mirror_binaries)
        self.assertTrue(repository.directories[Path("bin2/")].mirror_source)

    def test_broken(self):
        with self.assertRaises(RepositoryConfigException):
            self.get_config("BrokenConfig")

    def test_broken_flat(self):
        with self.assertRaises(RepositoryConfigException):
            self.get_config("FlatBrokenConfig")

    def test_netrc1(self):
        netrc = NetRC(self.TEST_DATA / "NetRC" / "auth1.conf")
        self.assertEqual(
            netrc.match_machine(URL.from_string("https://host1.example.com")),
            ("user", "password"),
        )

        netrc = NetRC(self.TEST_DATA / "NetRC" / "auth4.conf")
        self.assertEqual(
            netrc.match_machine(URL.from_string("http://host1.example.com/path123")),
            ("login1", "password1"),
        )

    def test_netrc2(self):
        def ensure_auth(
            config: Config,
            repository_url: str,
            login: str | None = None,
            password: str | None = None,
        ):
            repository = self.ensure_repository(config.repositories[repository_url])
            self.assertEqual(repository.url.username, login)
            self.assertEqual(repository.url.password, password)

        etc_netrc = self.TEST_DATA / "NetRC" / "auth1.conf"
        config = self.get_modified_config("NetRC", f"set etc_netrc {etc_netrc}")

        ensure_auth(config, "http://host1.example.com/path1")
        ensure_auth(config, "https://host1.example.com/path11", "user", "password")
        ensure_auth(config, "https://host2.example.com/path2")
        ensure_auth(config, "ftp://host3.example.com/path3")

        etc_netrc = self.TEST_DATA / "NetRC" / "auth2.conf"
        config = self.get_modified_config("NetRC", f"set etc_netrc {etc_netrc}")

        ensure_auth(config, "http://host1.example.com/path1")
        ensure_auth(config, "https://host1.example.com/path11")
        ensure_auth(config, "https://host2.example.com/path2")
        ensure_auth(config, "ftp://host3.example.com/path3", "ftp", "anon")

        etc_netrc = self.TEST_DATA / "NetRC" / "auth3.conf"
        config = self.get_modified_config("NetRC", f"set etc_netrc {etc_netrc}")

        ensure_auth(config, "http://host1.example.com/path1")
        ensure_auth(config, "https://host1.example.com/path11", "some", "thing")
        ensure_auth(config, "https://host2.example.com/path2")
        ensure_auth(config, "ftp://host3.example.com/path3")

        etc_netrc = self.TEST_DATA / "NetRC" / "auth4.conf"
        config = self.get_modified_config("NetRC", f"set etc_netrc {etc_netrc}")

        ensure_auth(config, "http://host1.example.com/path1", "login1", "password1")
        ensure_auth(config, "https://host1.example.com/path11", "login2", "password2")
        ensure_auth(config, "https://host2.example.com/path2")
        ensure_auth(config, "ftp://host3.example.com/path3")

    def test_slash_independent(self):
        config = self.get_config("SlashConfig")

        repository = self.ensure_repository(
            config.repositories["http://example.com/debian1"]
        )

        self.assertTrue(repository.clean)
        self.assertTrue(repository.mirror_dist_upgrader)
        self.assertEqual(repository.mirror_path, Path("path1"))
        self.assertIn("isource1", repository.package_filter.include_source_name)
        self.assertIn("esource1", repository.package_filter.exclude_source_name)
        self.assertIn("ibinary1", repository.package_filter.include_binary_packages)
        self.assertIn("ebinary1", repository.package_filter.exclude_binary_packages)

        repository = self.ensure_repository(
            config.repositories["http://example.com/debian2/"]
        )

        self.assertTrue(repository.clean)
        self.assertTrue(repository.mirror_dist_upgrader)
        self.assertEqual(repository.mirror_path, Path("path2"))
        self.assertIn("isource2", repository.package_filter.include_source_name)
        self.assertIn("esource2", repository.package_filter.exclude_source_name)
        self.assertIn("ibinary2", repository.package_filter.include_binary_packages)
        self.assertIn("ebinary2", repository.package_filter.exclude_binary_packages)

    def test_redundant_sources(self):
        config = self.get_config("MixedConfig")

        debian_security = self.ensure_repository(
            config.repositories["http://ftp.debian.org/debian-security"]
        )

        self.assertCountEqual(
            [str(s) for s in debian_security.sources_files],
            (
                "dists/bullseye-security/main/source/Sources",
                "dists/bullseye-security/contrib/source/Sources",
                "dists/bullseye-security/non-free-firmware/source/Sources",
                "dists/bullseye-security/non-free/source/Sources",
                "dists/bullseye-security/main/debian-installer/source/Sources",
            ),
        )

    def test_redundant_packages(self):
        config = self.get_config("MixedConfig")

        ubuntu = self.ensure_repository(
            config.repositories["http://archive.ubuntu.com/ubuntu"]
        )

        self.assertCountEqual(
            [str(s) for s in ubuntu.packages_files],
            (
                "dists/mantic/main/binary-amd64/Packages",
                "dists/mantic/main/binary-all/Packages",
                "dists/mantic/restricted/binary-amd64/Packages",
                "dists/mantic/restricted/binary-all/Packages",
                "dists/mantic/universe/binary-amd64/Packages",
                "dists/mantic/universe/binary-all/Packages",
                "dists/mantic/multiverse/binary-amd64/Packages",
                "dists/mantic/multiverse/binary-all/Packages",
                "dists/mantic-security/main/binary-amd64/Packages",
                "dists/mantic-security/main/binary-all/Packages",
                "dists/mantic-security/restricted/binary-amd64/Packages",
                "dists/mantic-security/restricted/binary-all/Packages",
                "dists/mantic-security/universe/binary-amd64/Packages",
                "dists/mantic-security/universe/binary-all/Packages",
                "dists/mantic-security/multiverse/binary-amd64/Packages",
                "dists/mantic-security/multiverse/binary-all/Packages",
                "dists/mantic-updates/main/binary-amd64/Packages",
                "dists/mantic-updates/main/binary-all/Packages",
                "dists/mantic-updates/restricted/binary-amd64/Packages",
                "dists/mantic-updates/restricted/binary-all/Packages",
                "dists/mantic-updates/universe/binary-amd64/Packages",
                "dists/mantic-updates/universe/binary-all/Packages",
                "dists/mantic-updates/multiverse/binary-amd64/Packages",
                "dists/mantic-updates/multiverse/binary-all/Packages",
                "dists/mantic-backports/main/binary-amd64/Packages",
                "dists/mantic-backports/main/binary-all/Packages",
                "dists/mantic-backports/restricted/binary-amd64/Packages",
                "dists/mantic-backports/restricted/binary-all/Packages",
                "dists/mantic-backports/universe/binary-amd64/Packages",
                "dists/mantic-backports/universe/binary-all/Packages",
                "dists/mantic-backports/multiverse/binary-amd64/Packages",
                "dists/mantic-backports/multiverse/binary-all/Packages",
                "dists/noble/main/binary-amd64/Packages",
                "dists/noble/main/binary-all/Packages",
                "dists/noble/restricted/binary-amd64/Packages",
                "dists/noble/restricted/binary-all/Packages",
                "dists/noble/universe/binary-amd64/Packages",
                "dists/noble/universe/binary-all/Packages",
                "dists/noble/multiverse/binary-amd64/Packages",
                "dists/noble/multiverse/binary-all/Packages",
                "dists/noble-security/main/binary-amd64/Packages",
                "dists/noble-security/main/binary-all/Packages",
                "dists/noble-security/restricted/binary-amd64/Packages",
                "dists/noble-security/restricted/binary-all/Packages",
                "dists/noble-security/universe/binary-amd64/Packages",
                "dists/noble-security/universe/binary-all/Packages",
                "dists/noble-security/multiverse/binary-amd64/Packages",
                "dists/noble-security/multiverse/binary-all/Packages",
                "dists/noble-updates/main/binary-amd64/Packages",
                "dists/noble-updates/main/binary-all/Packages",
                "dists/noble-updates/restricted/binary-amd64/Packages",
                "dists/noble-updates/restricted/binary-all/Packages",
                "dists/noble-updates/universe/binary-amd64/Packages",
                "dists/noble-updates/universe/binary-all/Packages",
                "dists/noble-updates/multiverse/binary-amd64/Packages",
                "dists/noble-updates/multiverse/binary-all/Packages",
                "dists/noble-backports/main/binary-amd64/Packages",
                "dists/noble-backports/main/binary-all/Packages",
                "dists/noble-backports/restricted/binary-amd64/Packages",
                "dists/noble-backports/restricted/binary-all/Packages",
                "dists/noble-backports/universe/binary-amd64/Packages",
                "dists/noble-backports/universe/binary-all/Packages",
                "dists/noble-backports/multiverse/binary-amd64/Packages",
                "dists/noble-backports/multiverse/binary-all/Packages",
            ),
        )



# SPDX-License-Identifer: GPL-3.0-or-later

from importlib.metadata import PackageNotFoundError, version

try:
    # We use plain string here instead of `__package__` to make pyinstaller
    # distribution collection works.
    __version__ = version("apt_mirror")
except PackageNotFoundError:
    __version__ = "unknown"



# SPDX-License-Identifer: GPL-3.0-or-later

import bz2
import gzip
import itertools
import lzma
import os
import shutil
from abc import ABC, abstractmethod
from collections.abc import Iterable, Sequence
from dataclasses import dataclass
from enum import Enum
from mmap import MADV_SEQUENTIAL, MAP_PRIVATE, mmap
from pathlib import Path
from typing import IO, Any

from debian.deb822 import Release

from .download import URL, DownloadFile, FileCompression, HashSum, HashType
from .filter import PackageFilter
from .logs import LoggerFactory

# https://github.com/pypy/pypy/issues/4991
# TODO: replace again with `from mmap import MAP_POPULATE`
MAP_POPULATE = 0x008000


def should_ignore_errors(ignored_paths: set[str], path: Path):
    return any(path.is_relative_to(ignored_path) for ignored_path in ignored_paths)


def is_safe_path(root_path: Path, path: Path):
    return (root_path / path).resolve().is_relative_to(root_path.resolve())


class InvalidReleaseFilesException(RuntimeError):
    def __init__(self, message: str) -> None:
        super().__init__(message)
        self.message = message


@dataclass
class PackageFile:
    path: Path
    size: int
    hashes: dict[HashType, HashSum]

    def to_download_file(self, directory: Path):
        download_file = DownloadFile.from_path(
            path=directory / self.path, check_size=True
        )
        for hash_type, hashsum in self.hashes.items():
            download_file.add_compression_variant(
                directory / self.path,
                size=self.size,
                hash_type=hash_type,
                hash_sum=hashsum,
                use_by_hash=False,
            )

        return download_file


class IndexFileParser(ABC):
    def __init__(
        self,
        repository_path: Path,
        index_files: set[Path],
        ignore_errors: set[str],
        logger_id: Any | None = None,
    ) -> None:
        super().__init__()

        self._log = LoggerFactory.get_logger(self, logger_id=logger_id)
        self._repository_path = repository_path
        self._index_files = index_files
        self._ignore_errors = ignore_errors
        self._pool_files: dict[Path, DownloadFile] = {}

    def parse(self) -> set[DownloadFile]:
        for index_file_relative_path in self._index_files:
            index_file = self._repository_path / index_file_relative_path

            if not self._unpack_index(index_file):
                if "binary-all" not in str(index_file):
                    self._log.warning(
                        f"Unable to unpack index file {index_file}. Skipping"
                    )
                continue

            index_file_size = index_file.stat().st_size
            if index_file_size < 1:
                continue

            with open(index_file, "rb") as fp:
                mfp = fp
                if index_file_size > 1 * 1024 * 1024:
                    mfp = mmap(
                        fp.fileno(),
                        length=0,
                        flags=MAP_PRIVATE | MAP_POPULATE,
                    )
                    mfp.madvise(MADV_SEQUENTIAL)

                try:
                    self._do_parse_index(mfp)
                finally:
                    if isinstance(mfp, mmap):
                        mfp.close()

        return set(self._pool_files.values())

    def _is_safe_path(self, path: Path):
        return is_safe_path(self._repository_path, path)

    def _unpack_index(self, file: Path) -> bool:
        for suffix, open_function in BaseRepository.COMPRESSION_SUFFIXES.items():
            compressed_file = file.with_name(f"{file.name}{suffix}")
            if not compressed_file.exists():
                continue

            try:
                with (
                    open_function(compressed_file, "rb") as source_fp,
                    open(file, "wb") as target_fp,
                ):
                    shutil.copyfileobj(source_fp, target_fp)
                    shutil.copystat(compressed_file, file)

                    return True
            except (lzma.LZMAError, OSError):
                return False

        return file.exists()

    def _should_ignore_errors(self, path: Path):
        return should_ignore_errors(self._ignore_errors, path)

    @abstractmethod
    def _do_parse_index(self, fp: IO[bytes] | mmap): ...


class SourcesParser(IndexFileParser):
    def __init__(
        self,
        repository_path: Path,
        index_files: set[Path],
        ignore_errors: set[str],
        package_filter: PackageFilter,
        logger_id: Any | None = None,
    ) -> None:
        super().__init__(
            repository_path, index_files, ignore_errors, logger_id=logger_id
        )

        self._package_filter = package_filter

        self._reset_block_parser()

    # https://github.com/pylint-dev/pylint/issues/5214
    def _reset_block_parser(self):
        self._package: str | None = None
        self._directory: Path | None = None
        self._hash_type = None
        self._package_files: dict[Path, PackageFile] = {}

    def _do_parse_index(self, fp: IO[bytes] | mmap):
        self._reset_block_parser()

        for bytes_line in itertools.chain(iter(fp.readline, b""), (b"\n",)):
            if bytes_line[0] == ord(" "):
                if not self._hash_type:
                    continue

                try:
                    hashsum, _size, filename = (
                        bytes_line.decode().strip().split(maxsplit=2)
                    )
                except ValueError:
                    continue

                if " " in filename:
                    continue

                file_path = Path(filename)
                if not self._is_safe_path(file_path):
                    self._log.warning(f"Skipping unsafe path: {file_path}")
                    continue

                self._package_files.setdefault(
                    file_path,
                    PackageFile(path=file_path, size=int(_size), hashes={}),
                ).hashes[self._hash_type] = HashSum(type=self._hash_type, hash=hashsum)
            elif bytes_line[0] != ord("\n"):
                match bytes_line:
                    case line if line.startswith(b"Package:"):
                        _, self._package = (  # pylint: disable=W0201
                            line.decode().strip().split()
                        )
                    case line if line.startswith(b"Directory:"):
                        # https://github.com/pylint-dev/pylint/issues/5214
                        _, directory = (  # pylint: disable=W0201
                            line.decode().strip().split()
                        )

                        directory = Path(directory)
                        if not self._is_safe_path(directory):
                            self._log.warning(f"Skipping unsafe Directory: {directory}")
                            continue

                        self._directory = directory  # pylint: disable=W0201
                    case line if line.startswith(b"Files:"):
                        self._hash_type = HashType.MD5  # pylint: disable=W0201
                        continue
                    case line if line.startswith(b"Checksums-"):
                        match line[len(b"Checksums-") : -1]:
                            case b"Sha1:":
                                self._hash_type = HashType.SHA1  # pylint: disable=W0201
                            case b"Sha256:":
                                self._hash_type = (  # pylint: disable=W0201
                                    HashType.SHA256
                                )
                            case b"Sha512:":
                                self._hash_type = (  # pylint: disable=W0201
                                    HashType.SHA512
                                )
                            case _:
                                self._hash_type = None  # pylint: disable=W0201

                        continue
                    case _:
                        self._hash_type = None  # pylint: disable=W0201
                        continue
            else:
                if not self._package or not self._directory:
                    self._reset_block_parser()
                    continue

                if not self._package_filter.package_allowed(self._package):
                    self._reset_block_parser()
                    continue

                for package_file in self._package_files.values():
                    download_file = package_file.to_download_file(self._directory)
                    download_file.ignore_errors = self._should_ignore_errors(
                        download_file.path
                    )
                    self._pool_files[download_file.path] = download_file

                self._reset_block_parser()


class PackagesParser(IndexFileParser):
    def __init__(
        self,
        repository_path: Path,
        index_files: set[Path],
        ignore_errors: set[str],
        package_filter: PackageFilter,
        logger_id: Any | None = None,
    ) -> None:
        super().__init__(
            repository_path, index_files, ignore_errors, logger_id=logger_id
        )

        self._package_filter = package_filter

        self._reset_block_parser()

    def _reset_block_parser(self):
        self._package = None
        self._source = None
        self._file_path = None
        self._size = 0
        self._hashes: dict[HashType, HashSum] = {}

    def _do_parse_index(self, fp: IO[bytes] | mmap):
        for bytes_line in itertools.chain(iter(fp.readline, b""), (b"\n",)):
            if bytes_line[0] != ord("\n"):
                match bytes_line:
                    case line if line.startswith(b"Package:"):
                        self._package = self._get_line_value(  # pylint: disable=W0201
                            line
                        )
                    case line if line.startswith(b"Source:"):
                        self._source = self._get_line_value(  # pylint: disable=W0201
                            line
                        ).split()[0]
                    case line if line.startswith(b"Filename:"):
                        file_path = Path(self._get_line_value(line))

                        if not self._is_safe_path(file_path):
                            self._log.warning(f"Skipping unsafe path: {file_path}")
                            continue

                        self._file_path = file_path  # pylint: disable=W0201
                    case line if line.startswith(b"Size:"):
                        self._size = int(  # pylint: disable=W0201
                            self._get_line_value(line)
                        )
                    case line if line.startswith(b"%s:" % HashType.MD5.value.encode()):
                        self._hashes[HashType.MD5] = HashSum(
                            type=HashType.MD5, hash=self._get_line_value(line)
                        )
                    case line if line.startswith(b"%s:" % HashType.SHA1.value.encode()):
                        self._hashes[HashType.MD5] = HashSum(
                            type=HashType.SHA1, hash=self._get_line_value(line)
                        )
                    case line if line.startswith(
                        b"%s:" % HashType.SHA256.value.encode()
                    ):
                        self._hashes[HashType.SHA256] = HashSum(
                            type=HashType.SHA256, hash=self._get_line_value(line)
                        )
                    case line if line.startswith(
                        b"%s:" % HashType.SHA512.value.encode()
                    ):
                        self._hashes[HashType.SHA512] = HashSum(
                            type=HashType.SHA512, hash=self._get_line_value(line)
                        )
                    case _:
                        self._hash_type = None  # pylint: disable=W0201
                        continue
            else:
                if not self._package or not self._file_path or not self._size:
                    self._reset_block_parser()
                    continue

                source_name = self._source if self._source else self._package

                if not self._package_filter.package_allowed(source_name, self._package):
                    self._reset_block_parser()
                    continue

                download_file = DownloadFile.from_path(self._file_path, check_size=True)
                download_file.ignore_errors = self._should_ignore_errors(
                    download_file.path
                )
                for hash_type, hashsum in self._hashes.items():
                    download_file.add_compression_variant(
                        path=self._file_path,
                        size=self._size,
                        hash_type=hash_type,
                        hash_sum=hashsum,
                        use_by_hash=False,
                    )

                self._pool_files[self._file_path] = download_file

                self._reset_block_parser()

    def _get_line_value(self, line: bytes):
        return line.decode().strip().split(":", maxsplit=1)[1].strip()


class ByHash(Enum):
    YES = "yes"
    NO = "no"
    FORCE = "force"

    @staticmethod
    def default():
        return ByHash.YES


@dataclass
class BaseRepositoryMetadata(ABC):
    by_hash: ByHash

    @abstractmethod
    def as_path(self) -> Path: ...

    def as_string(self) -> str:
        return str(self)

    def __hash__(self) -> int:
        return hash(str(self))


@dataclass(eq=False)
class Codename(BaseRepositoryMetadata):
    @dataclass
    class Component:
        name: str
        mirror_source: bool
        arches: list[str]

    codename: str
    components: dict[str, Component]

    def as_path(self) -> Path:
        return Path("dists") / self.codename

    def should_mirror_source(self):
        return any(c.mirror_source for c in self.components.values())

    def should_mirror_binaries(self):
        return any(c.arches for c in self.components.values())

    def __str__(self) -> str:
        return self.codename


@dataclass(eq=False)
class FlatDirectory(BaseRepositoryMetadata):
    directory: Path
    mirror_source: bool
    mirror_binaries: bool

    def as_path(self) -> Path:
        return self.directory

    def __str__(self) -> str:
        return str(self.directory)


@dataclass
class BaseRepository(ABC):
    COMPRESSION_SUFFIXES = {
        FileCompression.XZ.file_extension: lzma.open,
        FileCompression.GZ.file_extension: gzip.open,
        FileCompression.BZ2.file_extension: bz2.open,
    }

    ALL_INDEX_SUFFIXES = list(COMPRESSION_SUFFIXES.keys()) + [""]

    RELEASE_FILES = (
        "InRelease",
        "Release",
        "Release.gpg",
    )

    url: URL

    clean: bool
    skip_clean: set[Path]
    http2_disable: bool
    mirror_dist_upgrader: bool
    mirror_path: Path | None
    ignore_errors: set[str]

    def __post_init__(self):
        self._log = LoggerFactory.get_logger(self)

        self.package_filter = PackageFilter()

    def get_mirror_path(self, encode_tilde: bool):
        if self.mirror_path:
            return self.mirror_path

        return self.url.as_filesystem_path(encode_tilde)

    def as_filename(self, encode_tilde: bool) -> Path:
        return Path(
            str(self.get_mirror_path(encode_tilde=encode_tilde)).replace(os.sep, "_")
        )

    def get_clean_script_name(self, encode_tilde: bool):
        path = Path(str(self.get_mirror_path(encode_tilde)).replace("/", "_"))
        return path.with_name(f"{path.name}.sh")

    def get_metadata_files(
        self, repository_root: Path, encode_tilde: bool, missing_sources: set[Path]
    ) -> set[DownloadFile]:
        metadata_files: set[DownloadFile] = set()
        mirror_path = repository_root / self.get_mirror_path(encode_tilde)

        for (
            metadata,
            metadata_release_files,
        ) in self.release_files_per_metadata.items():
            codename_metadata_files: dict[Path, DownloadFile] = {}

            for release_file_relative_path in metadata_release_files:
                if release_file_relative_path.suffix == ".gpg":
                    continue

                if release_file_relative_path in missing_sources:
                    continue

                release_file = mirror_path / release_file_relative_path

                if not release_file.exists():
                    continue

                with open(release_file, "rt", encoding="utf-8") as fp:
                    release = Release(fp)

                use_hash = release.get("Acquire-By-Hash") == "yes"
                if use_hash:
                    if self.get_by_hash_policy(metadata) == ByHash.NO:
                        use_hash = False
                elif self.get_by_hash_policy(metadata) == ByHash.FORCE:
                    use_hash = True

                for hash_type in HashType:
                    for file in release.get(hash_type.value, []):
                        path = Path(file["name"])

                        if not is_safe_path(mirror_path, path):
                            self._log.warning(f"Skipping unsafe path: {path}")
                            continue

                        try:
                            size = int(file["size"])
                        except ValueError:
                            size = 0

                        if size <= 0:
                            continue

                        # Ignore release files in release files
                        if file["name"] in self.RELEASE_FILES:
                            continue

                        if not self._metadata_file_allowed(metadata, path):
                            continue

                        hash_sum = HashSum(type=hash_type, hash=file[hash_type.value])

                        repository_path = release_file_relative_path.parent / path
                        uncompressed_path = DownloadFile.uncompressed_path(path)
                        if uncompressed_path in codename_metadata_files:
                            codename_metadata_files[
                                uncompressed_path
                            ].add_compression_variant(
                                path=repository_path,
                                size=size,
                                hash_type=hash_type,
                                hash_sum=hash_sum,
                                use_by_hash=use_hash,
                            )
                        else:
                            codename_metadata_files[uncompressed_path] = (
                                DownloadFile.from_hashed_path(
                                    repository_path,
                                    size=size,
                                    hash_type=hash_type,
                                    hash_sum=hash_sum,
                                    use_by_hash=use_hash,
                                )
                            )

                        codename_metadata_files[
                            uncompressed_path
                        ].ignore_errors = should_ignore_errors(
                            self.ignore_errors, uncompressed_path
                        )

            if not codename_metadata_files:
                self._log.warning(f"No metadata files found for codename {metadata}")

            metadata_files.update(codename_metadata_files.values())

        return metadata_files

    def _get_sources_files_field(self, hash_type: HashType):
        if hash_type == HashType.MD5:
            return "Files"

        return f"Checksum-{hash_type.value.capitalize()}"

    def get_pool_files(
        self, repository_root: Path, encode_tilde: bool, missing_sources: set[Path]
    ) -> Iterable[DownloadFile]:
        pool_files: set[DownloadFile] = set()

        if self.is_source_enabled:
            pool_files.update(
                SourcesParser(
                    repository_root / self.get_mirror_path(encode_tilde),
                    set(self.sources_files) - missing_sources,
                    ignore_errors=self.ignore_errors,
                    package_filter=self.package_filter,
                    logger_id=self.url,
                ).parse()
            )

        if self.is_binaries_enabled:
            pool_files.update(
                PackagesParser(
                    repository_root / self.get_mirror_path(encode_tilde),
                    set(self.packages_files) - missing_sources,
                    ignore_errors=self.ignore_errors,
                    package_filter=self.package_filter,
                    logger_id=self.url,
                ).parse()
            )

        return pool_files

    def validate_release_files(self, repository_root: Path, encode_tilde: bool):
        release_files_exists = False

        for _, metadata_release_files in self.release_files_per_metadata.items():
            metadata_sizes: dict[str, list[tuple[int, Path]]] = {}
            metadata_hashes: dict[str, dict[HashType, list[tuple[str, Path]]]] = {}

            for release_file_relative_path in metadata_release_files:
                if release_file_relative_path.suffix == ".gpg":
                    continue

                release_file = (
                    repository_root
                    / self.get_mirror_path(encode_tilde)
                    / release_file_relative_path
                )

                if not release_file.exists():
                    continue

                with open(release_file, "rt", encoding="utf-8") as fp:
                    release = Release(fp)

                release_files_exists = True
                for hash_type in HashType:
                    for file in release.get(hash_type.value, []):
                        try:
                            size = int(file["size"])
                        except ValueError:
                            size = 0

                        if size <= 0:
                            continue

                        # Ignore release files in release files
                        if file["name"] in self.RELEASE_FILES:
                            continue

                        path = file["name"]
                        hash_sum = file[hash_type.value]

                        if any(size != s[0] for s in metadata_sizes.get(path, [])):
                            raise InvalidReleaseFilesException(
                                f"Size of file {path} in release file"
                                f" {release_file_relative_path} differs from size in"
                                f" release file {metadata_sizes[path][0][1]}"
                            )

                        if any(
                            hash_sum != s[0]
                            for s in metadata_hashes.get(path, {}).get(hash_type, [])
                        ):
                            raise InvalidReleaseFilesException(
                                f"Hashsum of type {hash_type} of file {path} in release"
                                f" file {release_file_relative_path} differs from"
                                " hashsum in release file"
                                f" {metadata_hashes[path][hash_type][0][1]}"
                            )

                        metadata_sizes.setdefault(path, []).append(
                            (size, release_file_relative_path)
                        )
                        metadata_hashes.setdefault(path, {}).setdefault(
                            hash_type, []
                        ).append((hash_sum, release_file_relative_path))

        if not release_files_exists:
            raise InvalidReleaseFilesException("No release files were found")

    @property
    def release_files(self) -> Sequence[Path]:
        return [
            path
            for _, paths in self.release_files_per_metadata.items()
            for path in paths
        ]

    @abstractmethod
    def _metadata_file_allowed(
        self, metadata: BaseRepositoryMetadata, file_path: Path
    ) -> bool: ...

    @abstractmethod
    def get_by_hash_policy(self, metadata: BaseRepositoryMetadata) -> ByHash: ...

    @property
    @abstractmethod
    def is_source_enabled(self) -> bool: ...

    @property
    @abstractmethod
    def is_binaries_enabled(self) -> bool: ...

    @property
    @abstractmethod
    def release_files_per_metadata(
        self,
    ) -> dict[BaseRepositoryMetadata, Sequence[Path]]: ...

    @property
    @abstractmethod
    def sources_files(self) -> Sequence[Path]: ...

    @property
    @abstractmethod
    def packages_files(self) -> Sequence[Path]: ...


@dataclass
class Repository(BaseRepository):
    class Codenames(dict[str, Codename]):
        def get_codename(self, codename: str) -> Codename:
            if codename not in self:
                raise RuntimeError(f"Requested codename was not found: {codename}")

            return self[codename]

        def __str__(self) -> str:
            return str([c for c in self])

    DISTS = Path("dists")
    DIST_UPGRADER_ANNOUNCEMENTS = [
        "DevelReleaseAnnouncement",
        "EOLReleaseAnnouncement",
        "ReleaseAnnouncement",
    ]

    codenames: Codenames

    def get_metadata_files(
        self, repository_root: Path, encode_tilde: bool, missing_sources: set[Path]
    ) -> set[DownloadFile]:
        metadata_files = super().get_metadata_files(
            repository_root, encode_tilde, missing_sources
        )

        if not self.mirror_dist_upgrader:
            return metadata_files

        # Add Ubuntu's dist-upgrader paths
        for codename in self.codenames.values():
            if "main" not in codename.components:
                continue

            dist_upgrader_path = Path(
                f"dists/{codename.codename}/main/dist-upgrader-all/current"
            )
            for file in self.DIST_UPGRADER_ANNOUNCEMENTS:
                announcement_path = dist_upgrader_path / file
                metadata_files.add(
                    DownloadFile.from_path(announcement_path, False, True)
                )
                metadata_files.add(
                    DownloadFile.from_path(
                        announcement_path.with_suffix(".html"), False, True
                    )
                )

            metadata_filename = codename.codename.split("-", maxsplit=1)[0]
            metadata_files.add(
                DownloadFile.from_path(
                    dist_upgrader_path / f"{metadata_filename}.tar.gz", False, True
                )
            )
            metadata_files.add(
                DownloadFile.from_path(
                    dist_upgrader_path / f"{metadata_filename}.tar.gz.gpg", False, True
                )
            )

        return metadata_files

    def _metadata_file_allowed(
        self, metadata: BaseRepositoryMetadata, file_path: Path
    ) -> bool:
        codename = self.codenames.get_codename(metadata.as_string())
        file_path_str = str(file_path)

        # Skip source metadata if not needed
        if not codename.should_mirror_source() and (
            "/source/" in file_path_str or file_path.name.startswith("Contents-source")
        ):
            return False

        # Skip binary metadata if not needed
        if not codename.should_mirror_binaries() and any(
            part in file_path_str for part in ("/binary-", "/cnf/", "/dep11/", "/i18n/")
        ):
            return False

        # Skip redundand components
        components_split = min(file_path_str.count("/"), 2)
        if components_split >= 1:
            file_component = file_path_str.rsplit("/", maxsplit=components_split)[0]

            if not any(
                file_component == component.name
                for component in codename.components.values()
            ):
                return False

            if (
                "/binary-" in file_path_str
                and "source" not in file_path_str
                and (
                    not codename.should_mirror_binaries()
                    or not any(
                        arch in file_path_str
                        for arch in itertools.chain(
                            codename.components[file_component].arches, ["-all"]
                        )
                    )
                )
            ):
                return False

        all_arches = set(
            arch
            for component in codename.components.values()
            for arch in component.arches
        )
        if all_arches:
            all_arches.add("all")

        if (
            any(
                file_path.name.startswith(suffix)
                for suffix in ("Commands-", "Components-", "Contents-")
            )
            and "source" not in file_path.name
            and not any(arch in file_path.name for arch in all_arches)
        ):
            return False

        if (
            "Contents-" in file_path_str
            and ".diff" in file_path_str
            and not any(arch in file_path.name for arch in all_arches)
        ):
            return False

        # Allow any metadata not explicitly disallowed
        return True

    @property
    def is_source_enabled(self) -> bool:
        return any(c.should_mirror_source() for c in self.codenames.values())

    @property
    def is_binaries_enabled(self) -> bool:
        return any(c.should_mirror_binaries() for c in self.codenames.values())

    @property
    def release_files_per_metadata(
        self,
    ) -> dict[BaseRepositoryMetadata, Sequence[Path]]:
        return {
            codename: [
                self.DISTS / codename.as_string() / file for file in self.RELEASE_FILES
            ]
            for codename in self.codenames.values()
        }

    @property
    def sources_files(self) -> Sequence[Path]:
        return [
            self.DISTS / codename.as_string() / component.name / "source" / "Sources"
            for codename in self.codenames.values()
            for component in codename.components.values()
            if component.mirror_source
        ]

    @property
    def packages_files(self) -> Sequence[Path]:
        return [
            path
            for path in itertools.chain.from_iterable(
                itertools.chain(
                    (
                        self.DISTS
                        / codename.as_string()
                        / component.name
                        / f"binary-{arch}"
                        / "Packages"
                        for arch in component.arches
                    ),
                    [
                        self.DISTS
                        / codename.as_string()
                        / component.name
                        / "binary-all"
                        / "Packages"
                    ],
                )
                for codename in self.codenames.values()
                for component in codename.components.values()
                if component.arches
            )
        ]

    def get_by_hash_policy(self, metadata: BaseRepositoryMetadata) -> ByHash:
        return self.codenames.get_codename(metadata.as_string()).by_hash

    def __str__(self) -> str:
        return (
            f"{self.url}, codenames: {self.codenames}, mirror_path: {self.mirror_path}"
        )


@dataclass
class FlatRepository(BaseRepository):
    class FlatDirectories(dict[Path, FlatDirectory]):
        def get_directory(self, directory: Path) -> FlatDirectory:
            if directory not in self:
                raise RuntimeError(f"Requested directory was not found: {directory}")

            return self[directory]

        def __str__(self) -> str:
            return str([d.as_string() for d in self.values()])

    directories: FlatDirectories

    def _metadata_file_allowed(
        self, metadata: BaseRepositoryMetadata, file_path: Path
    ) -> bool:
        directory = self.directories.get_directory(metadata.as_path())

        if not directory.mirror_source and (
            any(
                str(file_path) == f"Sources{suffix}"
                for suffix in self.ALL_INDEX_SUFFIXES
            )
        ):
            return False

        if not directory.mirror_binaries and (
            any(
                str(file_path) == f"Packages{suffix}"
                for suffix in self.ALL_INDEX_SUFFIXES
            )
        ):
            return False

        return True

    @property
    def is_source_enabled(self) -> bool:
        return any(d.mirror_source for d in self.directories.values())

    @property
    def is_binaries_enabled(self) -> bool:
        return any(d.mirror_binaries for d in self.directories.values())

    @property
    def release_files_per_metadata(
        self,
    ) -> dict[BaseRepositoryMetadata, Sequence[Path]]:
        return {
            directory: [directory.as_path() / file for file in self.RELEASE_FILES]
            for directory in self.directories.values()
        }

    @property
    def sources_files(self) -> Sequence[Path]:
        return [
            directory.as_path() / "Sources" for directory in self.directories.values()
        ]

    @property
    def packages_files(self) -> Sequence[Path]:
        return [
            directory.as_path() / "Packages" for directory in self.directories.values()
        ]

    def get_by_hash_policy(self, metadata: BaseRepositoryMetadata) -> ByHash:
        return self.directories.get_directory(metadata.as_path()).by_hash

    def __str__(self) -> str:
        return (
            f"{self.url}, directories: {self.directories},"
            f" mirror_path: {self.mirror_path}"
        )



from datetime import datetime
from pathlib import Path
from unittest.mock import patch

from apt_mirror.download import DownloadFile, HashSum, HashType
from apt_mirror.download.slow_rate_protector import SlowRateException, SlowRateProtector
from tests.base import BaseTest


class TestDownload(BaseTest):
    def test_download_file_size(self):
        file_path = Path("/tmp/t")
        file = DownloadFile.from_hashed_path(
            file_path, 1024, HashType.MD5, HashSum(HashType.MD5, ""), use_by_hash=True
        )
        file.add_compression_variant(file_path.with_suffix(".bz2"), 512)
        file.add_compression_variant(file_path.with_suffix(".xz"), 777)
        file.add_compression_variant(file_path.with_suffix(".gz"), 256)

        self.assertEqual(file.size, 777)

    def test_slow_rate_protector(self):
        with patch(f"{SlowRateProtector.__module__}.datetime") as mock_datetime:
            mock_datetime.now.return_value = datetime(2024, 1, 1, 0, 0, 0)

            protector = SlowRateProtector("test", 60, 100)
            protector.rate(1)

            mock_datetime.now.return_value = datetime(2024, 1, 1, 0, 0, 59)
            protector.rate(1)

            mock_datetime.now.return_value = datetime(2024, 1, 1, 0, 1, 0)
            with self.assertRaises(SlowRateException):
                protector.rate(1)

            with self.assertRaises(SlowRateException):
                protector.rate(100 * 60 - 4)

            protector.rate(1)



# SPDX-License-Identifer: GPL-3.0-or-later

import itertools
import re
from collections.abc import Generator
from dataclasses import dataclass
from enum import Enum
from pathlib import Path
from urllib.parse import urlparse

from .download.url import URL


class TokenType(Enum):
    MACHINE = "machine"
    LOGIN = "login"
    PASSWORD = "password"


@dataclass
class Token:
    token: TokenType
    value: str


@dataclass(frozen=True)
class Machine:
    protocol: str | None
    hostname: str | None
    port: int | None
    path: str | None

    @classmethod
    def from_string(cls, string: str):
        if "//" not in string:
            string = f"//{string}"

        url = urlparse(string)

        return cls(
            protocol=url.scheme, hostname=url.hostname, port=url.port, path=url.path
        )


class NetRCFile:
    def __init__(self, path: Path) -> None:
        self._path = path

    def tokens(self) -> Generator[Token, None, None]:
        if not self._path.is_file():
            return

        with open(self._path, "rt", encoding="utf-8") as fp:
            for line in fp:
                parts = (t for t in re.split(r"[ \t\n\r]", line) if t)
                try:
                    while True:
                        try:
                            token = TokenType(next(parts))
                        except ValueError:
                            next(parts)
                            continue

                        yield Token(token, next(parts))
                except (StopIteration, ValueError):
                    continue


class NetRC:
    def __init__(self, netrc_path: Path = Path("/etc/apt/auth.conf")) -> None:
        self._machines: dict[Machine, tuple[str, str]] = {}
        self._reset_machine()

        for file in itertools.chain(
            [netrc_path],
            netrc_path.with_name(f"{netrc_path.name}.d").glob("*.conf"),
        ):
            self._process_file(file)

    def _process_file(self, file: Path) -> None:
        for token in NetRCFile(file).tokens():
            match token.token:
                case TokenType.MACHINE:
                    self._add_machine()
                    self._machine = Machine.from_string(token.value)
                case TokenType.LOGIN:
                    self._login = token.value
                case TokenType.PASSWORD:
                    self._password = token.value

        self._add_machine()

    def _add_machine(self) -> None:
        if self._machine and self._login and self._password:
            self._machines[self._machine] = (self._login, self._password)

        self._reset_machine()

    def _reset_machine(self) -> None:
        self._machine: Machine | None = None
        self._login: str | None = None
        self._password: str | None = None

    def match_machine(self, url: URL) -> tuple[str, str] | None:
        for machine, auth in self._machines.items():
            if machine.protocol != url.scheme and (
                machine.protocol or url.scheme not in ("https", "tor+https")
            ):
                continue

            if machine.hostname != url.hostname:
                continue

            if machine.port and machine.port != url.port:
                continue

            if machine.path and not url.path.startswith(machine.path):
                continue

            return auth

        return None



# SPDX-License-Identifer: GPL-3.0-or-later

from abc import ABC, abstractmethod
from collections.abc import Generator, Iterable
from typing import Any

from .download.downloader import Downloader
from .repository import BaseRepository


class BaseDownloaderCollector(ABC):
    def __init__(self, address: str, port: int) -> None:
        self._address = address
        self._port = port
        self._repositories: list[tuple[BaseRepository, Downloader]] = []

    def prometheus_available(self) -> bool:
        return False

    def shutdown(self):  # noqa: B027
        pass

    def add_downloader(self, repository: BaseRepository, downloader: Downloader):
        self._repositories.append((repository, downloader))

    @abstractmethod
    def collect(self) -> Iterable[Any]:
        pass


class DummyDownloaderCollector(BaseDownloaderCollector):
    def collect(self):
        yield


try:
    from prometheus_client import Metric, start_http_server
    from prometheus_client.core import REGISTRY, GaugeMetricFamily
    from prometheus_client.registry import Collector
except ImportError:

    class DownloaderCollector(DummyDownloaderCollector):
        pass

else:

    class DownloaderCollector(BaseDownloaderCollector, Collector):  # type: ignore
        def __init__(self, address: str, port: int) -> None:
            super().__init__(address, port)

            self._wsgi_server = None
            self._wsgi_thread = None

            wsgi_data = start_http_server(port=port, addr=address)
            if wsgi_data:
                self._wsgi_server, self._wsgi_thread = wsgi_data

            REGISTRY.register(self)

        def prometheus_available(self) -> bool:
            return True

        def shutdown(self):
            if self._wsgi_server:
                self._wsgi_server.shutdown()

                if self._wsgi_thread:
                    self._wsgi_thread.join()

        def _metric(self, name: str):
            mf = GaugeMetricFamily(
                f"apt_mirror_{name}",
                name.replace("_", " ").capitalize(),
                labels=["url"],
            )

            for repository, downloader in self._repositories:
                mf.add_metric(
                    [str(repository.url)],
                    value=getattr(downloader, name),
                )

            return mf

        def collect(self) -> Generator[Metric, Any, None]:
            yield self._metric("queue_files_count")
            yield self._metric("queue_files_size")
            yield self._metric("downloaded_files_count")
            yield self._metric("downloaded_files_size")
            yield self._metric("error_files_count")
            yield self._metric("error_files_size")
            yield self._metric("missing_files_count")
            yield self._metric("missing_files_size")
            yield self._metric("unmodified_files_count")
            yield self._metric("unmodified_files_size")



# SPDX-License-Identifer: GPL-3.0-or-later

import contextlib
from abc import ABC, abstractmethod
from collections.abc import AsyncIterator
from contextlib import asynccontextmanager
from pathlib import Path
from typing import Protocol

from .logs import LoggerFactory


class AsyncSupportsWrite(Protocol):
    async def write(self, data: bytes) -> int: ...


class BaseAsyncIOFileWriterFactory(ABC):
    MODE = "wb"

    def __init__(self) -> None:
        self._log = LoggerFactory.get_logger(self)

    @classmethod
    async def create(cls, *test_paths: Path) -> "BaseAsyncIOFileWriterFactory":
        clazz = cls()
        await clazz.test_storage(*test_paths)

        return clazz

    @abstractmethod
    async def test_storage(self, *test_paths: Path): ...

    @asynccontextmanager
    @abstractmethod
    async def open(self, path: Path) -> AsyncIterator[AsyncSupportsWrite]:
        yield  # type: ignore


try:
    from aiofile import async_open as aiofile_open
    from caio import (
        AsyncioContext,
        linux_aio_asyncio,
        python_aio_asyncio,
        thread_aio_asyncio,
    )

    class AsyncIOFileFactory(BaseAsyncIOFileWriterFactory):  # type: ignore
        def __init__(self) -> None:
            super().__init__()
            self._context = self._get_supported_context()

        def _get_supported_context(
            self, fallback_context: bool = False
        ) -> AsyncioContext:
            if linux_aio_asyncio and not fallback_context:
                return linux_aio_asyncio.AsyncioContext()
            elif thread_aio_asyncio:
                if not fallback_context:
                    self._log.warning(
                        "Native Linux AIO doesn't supported on this system. "
                        "Fallback to threaded AIO implementation."
                    )
                return thread_aio_asyncio.AsyncioContext()
            else:
                if not fallback_context:
                    self._log.warning(
                        "Nor native Linux AIO nor threaded AIO implementation are "
                        "supported on this system. Fallback to pure Python "
                        "implementation."
                    )
                return python_aio_asyncio.AsyncioContext()

        @asynccontextmanager
        async def _open(
            self,
            path: Path,
        ):
            yield await aiofile_open(path, mode=self.MODE, context=self._context)

        async def test_storage(self, *test_paths: Path) -> None:
            for path in test_paths:
                try:
                    path.parent.mkdir(parents=True, exist_ok=True)
                    async with self._open(path) as fp:
                        await fp.write(b" ")

                    path.unlink()
                except SystemError as e:
                    if "not supported" in str(e):
                        self._context = self._get_supported_context(
                            fallback_context=True
                        )
                        self._log.warning(
                            f"Linux AIO check failed for path {path}. Falling back to"
                            " non-AIO IO implementation."
                        )
                        break
                finally:
                    with contextlib.suppress(OSError):
                        path.unlink(missing_ok=True)

        @asynccontextmanager
        async def open(self, path: Path) -> AsyncIterator[AsyncSupportsWrite]:
            async with self._open(path) as fd:
                yield fd

except ImportError:
    from aiofiles import open as aiofiles_open
    from aiofiles.threadpool.binary import AsyncBufferedIOBase

    LoggerFactory.get_logger(__package__).warning(
        "True AIO is disabled because `aiofile` is not available, falling back to"
        " threaded wrapper via `aiofiles`. Consider using aiofile instead."
    )

    class AIOFilesWriter:
        def __init__(self, fd: AsyncBufferedIOBase) -> None:
            self._fd = fd

        async def write(self, data: bytes) -> int:
            return await self._fd.write(data)

    class AsyncIOFileFactory(BaseAsyncIOFileWriterFactory):
        @asynccontextmanager
        async def open(self, path: Path) -> AsyncIterator[AsyncSupportsWrite]:
            async with aiofiles_open(path, self.MODE) as fd:
                yield AIOFilesWriter(fd)



# SPDX-License-Identifer: GPL-3.0-or-later

import asyncio
import contextlib
import itertools
import os
import shutil
from abc import ABC, abstractmethod
from collections.abc import AsyncGenerator
from contextlib import asynccontextmanager
from dataclasses import dataclass
from datetime import datetime, timezone
from pathlib import Path
from typing import Any

from aiolimiter import AsyncLimiter

from ..aiofile import BaseAsyncIOFileWriterFactory
from ..logs import LoggerFactory
from .download_file import DownloadFile, DownloadFileCompressionVariant
from .format import format_size
from .proxy import Proxy
from .response import DownloadResponse
from .slow_rate_protector import SlowRateProtectorFactory
from .url import URL


@dataclass
class DownloaderSettings:
    url: URL
    target_root_path: Path
    aiofile_factory: BaseAsyncIOFileWriterFactory
    proxy: Proxy
    http2_disable: bool
    user_agent: str
    semaphore: asyncio.Semaphore
    slow_rate_protector_factory: SlowRateProtectorFactory
    rate_limiter: AsyncLimiter | None = None
    verify_ca_certificate: bool | str = True
    client_certificate: str | None = None
    client_private_key: str | None = None


class Downloader(ABC):
    BUFFER_SIZE = 8 * 1024 * 1024

    def __init__(self, *, settings: DownloaderSettings):
        self._log = LoggerFactory.get_logger(
            self,
            logger_id=settings.url,
        )

        self._settings = settings

        # Download queue. Reseted in download()
        self._sources: list[DownloadFile] = []
        # Downloaded files
        self._downloaded: list[DownloadFileCompressionVariant] = []
        # Umnodified files
        self._unmodified: list[DownloadFileCompressionVariant] = []
        # Either missing on server files or files with errors
        self._missing_sources: set[Path] = set()
        self._download_start = datetime.now()

        self.reset_stats()

        self.__post_init__()

    def __post_init__(self):  # noqa: B027
        pass

    def reset_stats(self):
        self._downloaded_count = 0
        self._downloaded_size = 0
        self._unmodified_count = 0
        self._unmodified_size = 0
        self._missing_count = 0
        self._missing_size = 0
        self._error_count = 0
        self._error_size = 0

    def reset_paths(self):
        self._downloaded: list[DownloadFileCompressionVariant] = []
        self._unmodified: list[DownloadFileCompressionVariant] = []
        self._missing_sources: set[Path] = set()

    def set_target_path(self, path: Path):
        self._settings.target_root_path = path

    def add(self, *args: DownloadFile):
        self._sources.extend(a for a in args)

        self.reset_stats()

    @property
    def queue_files_count(self) -> int:
        return len(self._sources)

    @property
    def queue_files_size(self) -> int:
        return sum(file.size for file in self._sources)

    @property
    def queue_files_formatted_size(self) -> str:
        return format_size(self.queue_files_size)

    @property
    def downloaded_files_count(self) -> int:
        return self._downloaded_count

    @property
    def downloaded_files_size(self) -> int:
        return self._downloaded_size

    @property
    def error_files_count(self) -> int:
        return self._error_count

    @property
    def error_files_size(self) -> int:
        return self._error_size

    @property
    def missing_files_count(self) -> int:
        return self._missing_count

    @property
    def missing_files_size(self) -> int:
        return self._missing_size

    @property
    def unmodified_files_count(self) -> int:
        return self._unmodified_count

    @property
    def unmodified_files_size(self) -> int:
        return self._unmodified_size

    async def download(self):
        async def remove_finished_tasks(tasks: set[asyncio.Task[Any]]):
            done_tasks, _ = await asyncio.wait(
                tasks, return_when=asyncio.FIRST_COMPLETED
            )

            tasks.difference_update(done_tasks)

        self._download_start = datetime.now()
        tasks: set[asyncio.Task[Any]] = set()
        progress_task = asyncio.create_task(self.progress_logger())

        while self._sources:
            source_file = self._sources.pop()

            file_unmodified = False
            if source_file.check_size:
                for variant in source_file.iter_variants():
                    target_path = (
                        self._settings.target_root_path / variant.get_source_path()
                    )

                    try:
                        stat = target_path.stat()
                        if stat.st_size == source_file.size:
                            self._unmodified_count += 1
                            self._unmodified_size += source_file.size
                            self._unmodified.append(variant)

                            file_unmodified = True
                            break
                    except FileNotFoundError:
                        pass

            if file_unmodified:
                continue

            tasks.add(asyncio.create_task(self.download_file(source_file)))

            if len(tasks) >= 128:
                await remove_finished_tasks(tasks)

        while tasks:
            await remove_finished_tasks(tasks)

        progress_task.cancel()
        with contextlib.suppress(asyncio.CancelledError):
            await progress_task

        self.log_status("Download finished")

    def need_update(self, path: Path, size: int | None, date: datetime | None) -> bool:
        if path.exists() and date and size:
            stat = path.stat()
            target_date = datetime.fromtimestamp(stat.st_mtime, tz=timezone.utc)
            target_size = stat.st_size

            if date == target_date and size == target_size:
                return False

        return True

    async def progress_logger(self):
        while True:
            try:
                await asyncio.sleep(10)
            except asyncio.CancelledError:
                return

            self.log_status("Download progress")

    def log_status(self, message: str):
        download_rate = format_size(
            self._downloaded_size
            / (datetime.now().timestamp() - self._download_start.timestamp()),
            suffix="B/sec",
        )
        self._log.info(
            message
            + f": {self._downloaded_count} ({format_size(self._downloaded_size)},"
            f" {download_rate});"
            " unmodified:"
            f" {self._unmodified_count} ({format_size(self._unmodified_size)});"
            f" missing: {self._missing_count} ({format_size(self._missing_size)});"
            f" errors: {self._error_count} ({format_size(self._error_size)})"
        )

    async def download_file(self, source_file: DownloadFile):
        async def retry(
            message: str | None = None, sleep: bool = True, skip_try: bool = False
        ):
            nonlocal tries
            if message:
                self._log.warning(message)

            if not skip_try:
                tries -= 1

            if sleep:
                await asyncio.sleep(5)

        error = False
        for variant in source_file.iter_variants():
            expected_size = variant.size

            for source_path in variant.get_all_paths():
                target_path = self._settings.target_root_path / source_path

                tries = 10
                while tries > 0:
                    async with (
                        self._settings.semaphore,
                        self.stream(source_path) as response,
                    ):
                        target_path.parent.mkdir(parents=True, exist_ok=True)

                        if response.retry:
                            await retry(skip_try=True)
                            continue

                        if response.missing:
                            if source_file.ignore_errors or source_file.ignore_missing:
                                break

                            await retry(
                                f"File {source_path} is missing from server."
                                " Retrying..."
                            )
                            continue

                        if response.error:
                            if source_file.ignore_errors:
                                break

                            await retry(
                                f"Received error `{response.error}` while downloading"
                                f" {source_path}. Retrying..."
                            )
                            error = True
                            continue

                        if (
                            expected_size > 0
                            and response.size
                            and response.size > 0
                            and expected_size != response.size
                        ):
                            if source_file.ignore_errors:
                                break

                            await retry(
                                f"Server reported size {response.size} differs from"
                                f" expected size {expected_size} for file"
                                f" {source_path}. Retrying..."
                            )
                            error = True
                            continue

                        mirror_paths = [
                            self._settings.target_root_path / path
                            for path in variant.get_all_paths()
                        ]

                        if response.size and not self.need_update(
                            target_path, response.size, response.date
                        ):
                            self._unmodified_count += 1
                            self._unmodified_size += response.size

                            if mirror_paths:
                                self.link_or_copy(target_path, *mirror_paths)

                            self._downloaded.append(variant)
                            self._missing_sources.difference_update(
                                variant.get_all_paths()
                            )

                            return

                        size = 0
                        target_path.unlink(missing_ok=True)
                        async with self._settings.aiofile_factory.open(
                            target_path
                        ) as fp:
                            try:
                                slow_rate_protector_factory = (
                                    self._settings.slow_rate_protector_factory
                                )
                                slow_rate_protector = (
                                    slow_rate_protector_factory.for_target(
                                        variant.get_source_path()
                                    )
                                )
                                async for chunk in response.stream():
                                    if self._settings.rate_limiter:
                                        await self._settings.rate_limiter.acquire(
                                            min(
                                                len(chunk),
                                                self._settings.rate_limiter.max_rate,
                                            )
                                        )

                                    size += len(chunk)
                                    slow_rate_protector.rate(len(chunk))
                                    await fp.write(chunk)
                            except Exception as ex:  # pylint: disable=W0718
                                await retry(
                                    f"An error `{ex.__class__.__qualname__}: {ex}`"
                                    f" occured while downloading file {source_path}."
                                    " Retrying..."
                                )
                                error = True
                                continue

                        if expected_size > 0 and expected_size != size:
                            await retry(
                                f"Downloaded size {size} is differs from expected size"
                                f" {expected_size} for file {source_path}. Retrying..."
                            )
                            error = True
                            continue

                        if response.date:
                            os.utime(
                                target_path,
                                (response.date.timestamp(), response.date.timestamp()),
                            )

                        if mirror_paths:
                            self.link_or_copy(target_path, *mirror_paths)

                        self._downloaded_count += 1
                        self._downloaded_size += size

                        self._downloaded.append(variant)
                        self._missing_sources.difference_update(variant.get_all_paths())
                        return

        if source_file.ignore_errors:
            self._log.info(f"Unable to download `{source_file.path}`: ignoring")
            return

        if source_file.ignore_missing and not error:
            self._log.info(f"Optional file `{source_file.path}` is missing on server")
            return

        self._missing_sources.update(
            itertools.chain.from_iterable(
                v.get_all_paths() for v in source_file.compression_variants.values()
            )
        )

        if not error:
            self._missing_count += 1
            self._missing_size += source_file.size

            self._log.warning(
                f"Unable to download {source_file.path}: file is missing on server"
            )
            return

        self._error_count += 1
        self._error_size += source_file.size

        self._log.error(f"Unable to download {source_file.path}: no more tries")

    def get_downloaded_files(self) -> list[DownloadFileCompressionVariant]:
        return self._downloaded.copy()

    def get_unmodified_files(self) -> list[DownloadFileCompressionVariant]:
        return self._unmodified.copy()

    def get_all_files(self) -> list[DownloadFileCompressionVariant]:
        return self.get_downloaded_files() + self.get_unmodified_files()

    def get_downloaded_files_paths(self) -> set[Path]:
        return (
            set(
                itertools.chain.from_iterable(
                    v.get_all_paths() for v in self.get_all_files()
                )
            )
            - self.get_missing_sources()
        )

    def get_missing_sources(self):
        return self._missing_sources.copy()

    def has_errors(self):
        return self._error_count > 0

    def has_missing(self):
        return self._missing_count > 0

    @staticmethod
    def link_or_copy(source: Path, *targets: Path):
        if len(targets) > 1:
            Downloader.link_or_copy(source, targets[0])
            source = targets[0]

        for target in targets:
            if target == source:
                continue

            target.parent.mkdir(parents=True, exist_ok=True)
            target.unlink(missing_ok=True)

            try:
                target.hardlink_to(source)
            except OSError:
                shutil.copy2(source, target)

    @asynccontextmanager
    @abstractmethod
    async def stream(self, source_path: Path) -> AsyncGenerator[DownloadResponse, None]:
        yield  # type: ignore



from collections.abc import Iterable
from pathlib import Path
from tempfile import TemporaryDirectory
from unittest import TestCase
from unittest.mock import patch

from apt_mirror.download import DownloadFile
from apt_mirror.download.url import URL
from apt_mirror.filter import PackageFilter
from apt_mirror.repository import (
    ByHash,
    Codename,
    InvalidReleaseFilesException,
    PackagesParser,
    Repository,
    SourcesParser,
    is_safe_path,
)
from tests.base import BaseTest


class TestRepository(BaseTest):
    @staticmethod
    def get_from_set(collection: Iterable[DownloadFile], element: Path) -> DownloadFile:
        for e in collection:
            if hash(e) == hash(element):
                return e

        raise RuntimeError(f"No element {element} found in set {collection}")

    def get_repository(
        self,
        components: list[str] | None = None,
        arches: list[str] | None = None,
        mirror_source: bool = True,
        mirror_dist_upgrader: bool = False,
    ):
        if components is None:
            components = ["main", "contrib", "non-free", "non-free/debian-installer"]

        if arches is None:
            arches = ["amd64", "s390x"]

        return Repository(
            url=URL.from_string("http://localhost.local/repo"),
            clean=False,
            skip_clean=set(),
            http2_disable=False,
            mirror_dist_upgrader=mirror_dist_upgrader,
            mirror_path=Path("repo"),
            ignore_errors=set(),
            codenames=Repository.Codenames(
                [
                    (
                        "test",
                        Codename(
                            ByHash.default(),
                            "test",
                            {
                                component: Codename.Component(
                                    component, mirror_source, arches
                                )
                                for component in components
                            },
                        ),
                    )
                ]
            ),
        )

    def test_ignore_errors(self):
        config = self.get_config("IgnoreErrorsConfig")

        repository = self.ensure_repository(
            config.repositories[
                "https://packages.gitlab.com/runner/gitlab-runner/debian"
            ]
        )

        self.assertTrue(repository.ignore_errors)

        files = repository.get_pool_files(
            self.TEST_DATA / "IgnoreErrorsConfig", False, set()
        )

        file = self.get_from_set(
            files,
            Path("pool/bullseye/main/g/gitlab-runner/gitlab-runner_14.8.1_amd64.deb"),
        )
        self.assertTrue(file.ignore_errors)

        file = self.get_from_set(
            files,
            Path("pool/bullseye/main/g/gitlab-runner/gitlab-runner_16.8.0_amd64.deb"),
        )
        self.assertFalse(file.ignore_errors)

        repository = self.ensure_repository(
            config.repositories["http://ftp.debian.org/debian-security"]
        )
        self.assertFalse(repository.ignore_errors)

    def test_release_files(self):
        repository = self.get_repository()

        with self.assertRaises(InvalidReleaseFilesException):
            repository.validate_release_files(
                self.TEST_DATA / "UnsyncedReleaseFiles", False
            )

        repository.validate_release_files(self.TEST_DATA / "SyncedReleaseFiles", False)

        with self.assertRaises(InvalidReleaseFilesException):
            repository.validate_release_files(
                self.TEST_DATA / "NonExistingFolder", False
            )

    def test_components_filter_buster(self):
        repository = self.get_repository(
            components=["main", "contrib", "non-free", "non-free/debian-installer"],
            arches=["amd64", "s390x"],
            mirror_source=True,
        )

        metadata_files = set(
            d.path.relative_to(Path("dists/test"))
            for d in repository.get_metadata_files(
                self.TEST_DATA / "DebianArchiveBusterProposedUpdates", False, set()
            )
        )

        self.assertIn(Path("main/source/Sources"), metadata_files)
        self.assertIn(Path("main/binary-amd64/Packages"), metadata_files)
        self.assertIn(Path("contrib/binary-amd64/Packages"), metadata_files)
        self.assertIn(Path("contrib/i18n/Translation-en"), metadata_files)
        self.assertIn(Path("non-free/binary-all/Packages"), metadata_files)
        self.assertIn(
            Path("non-free/debian-installer/binary-all/Release"), metadata_files
        )

        self.assertNotIn(Path("main/Contents-ppc64el"), metadata_files)
        self.assertNotIn(Path("contrib/binary-arm64/Packages"), metadata_files)
        self.assertNotIn(Path("non-free-firmware/Contents-amd64"), metadata_files)

        repository = self.get_repository(
            components=["contrib"],
            arches=["i386"],
            mirror_source=False,
        )

        metadata_files = set(
            d.path.relative_to(Path("dists/test"))
            for d in repository.get_metadata_files(
                self.TEST_DATA / "DebianArchiveBusterProposedUpdates", False, set()
            )
        )

        self.assertNotIn(Path("main/source/Sources"), metadata_files)
        self.assertNotIn(Path("contrib/source/Sources"), metadata_files)

        self.assertIn(Path("contrib/binary-all/Packages"), metadata_files)

        self.assertNotIn(Path("main/binary-amd64/Packages"), metadata_files)
        self.assertIn(Path("contrib/binary-i386/Packages"), metadata_files)

    def test_components_filter_bookworm(self):
        repository = self.get_repository(
            components=["main", "contrib", "non-free", "non-free/debian-installer"],
            arches=["amd64", "s390x"],
            mirror_source=True,
        )

        metadata_files = set(
            d.path.relative_to(Path("dists/test"))
            for d in repository.get_metadata_files(
                self.TEST_DATA / "DebianBookworm", False, set()
            )
        )

        self.assertIn(Path("main/Contents-all"), metadata_files)
        self.assertIn(Path("main/Contents-udeb-all"), metadata_files)
        self.assertIn(Path("main/Contents-amd64"), metadata_files)
        self.assertIn(Path("main/Contents-udeb-amd64"), metadata_files)
        self.assertIn(Path("main/Contents-s390x"), metadata_files)
        self.assertIn(Path("main/Contents-udeb-s390x"), metadata_files)

        self.assertNotIn(Path("main/Contents-mips64el"), metadata_files)
        self.assertNotIn(Path("main/Contents-udeb-mips64el"), metadata_files)
        self.assertNotIn(Path("main/Contents-udeb-ppc64el"), metadata_files)
        self.assertNotIn(Path("main/Contents-ppc64el"), metadata_files)

    def test_components_filter_all(self):
        repository = self.get_repository(
            components=["main", "contrib", "non-free", "non-free/debian-installer"],
            arches=[],
            mirror_source=True,
        )

        metadata_files = set(
            d.path.relative_to(Path("dists/test"))
            for d in repository.get_metadata_files(
                self.TEST_DATA / "DebianBookworm", False, set()
            )
        )

        self.assertNotIn(Path("main/Contents-all"), metadata_files)
        self.assertNotIn(Path("main/Contents-udeb-all"), metadata_files)
        self.assertNotIn(Path("main/Contents-amd64"), metadata_files)
        self.assertNotIn(Path("main/Contents-udeb-amd64"), metadata_files)
        self.assertNotIn(Path("main/Contents-s390x"), metadata_files)
        self.assertNotIn(Path("main/Contents-udeb-s390x"), metadata_files)
        self.assertNotIn(Path("main/Contents-mips64el"), metadata_files)
        self.assertNotIn(Path("main/Contents-udeb-mips64el"), metadata_files)
        self.assertNotIn(Path("main/Contents-udeb-ppc64el"), metadata_files)
        self.assertNotIn(Path("main/Contents-ppc64el"), metadata_files)

        self.assertNotIn(Path("main/binary-all/Packages"), metadata_files)

    def test_release_gpg(self):
        repository = self.get_repository(
            components=["main", "contrib", "non-free", "non-free/debian-installer"],
            arches=[],
            mirror_source=True,
        )

        try:
            repository.get_metadata_files(
                self.TEST_DATA / "BinaryGPGReleaseFile", False, set()
            )
        except UnicodeDecodeError:
            self.fail(
                "get_metadata_files() must not raise UnicodeDecodeError in case"
                " Release.gpg is binary"
            )

    def test_include_source_name(self):
        config = self.get_config("DebianBookworm", config_name="mirror1.list")

        repository = self.ensure_repository(
            config.repositories["http://ftp.debian.org/debian"]
        )
        repository.mirror_path = Path("repo")

        files = {
            d.path
            for d in repository.get_pool_files(
                self.TEST_DATA / "DebianBookworm", False, set()
            )
        }

        self.assertEqual(
            {
                Path(p)
                for p in [
                    "pool/main/c/curl/curl_7.88.1-10+deb12u5.debian.tar.xz",
                    "pool/main/c/curl/curl_7.88.1-10+deb12u5.dsc",
                    "pool/main/c/curl/curl_7.88.1-10+deb12u5_amd64.deb",
                    "pool/main/c/curl/curl_7.88.1.orig.tar.gz",
                    "pool/main/c/curl/curl_7.88.1.orig.tar.gz.asc",
                    "pool/main/c/curl/libcurl3-gnutls_7.88.1-10+deb12u5_amd64.deb",
                    "pool/main/c/curl/libcurl3-nss_7.88.1-10+deb12u5_amd64.deb",
                    "pool/main/c/curl/libcurl4-doc_7.88.1-10+deb12u5_all.deb",
                    "pool/main/c/curl/libcurl4-gnutls-dev_7.88.1-10+deb12u5_amd64.deb",
                    "pool/main/c/curl/libcurl4-nss-dev_7.88.1-10+deb12u5_amd64.deb",
                    "pool/main/c/curl/libcurl4-openssl-dev_7.88.1-10+deb12u5_amd64.deb",
                    "pool/main/c/curl/libcurl4_7.88.1-10+deb12u5_amd64.deb",
                ]
            },
            files,
        )

    def test_exclude_source_name(self):
        config = self.get_config("DebianBookworm", config_name="mirror2.list")

        repository = self.ensure_repository(
            config.repositories["http://ftp.debian.org/debian"]
        )
        repository.mirror_path = Path("repo")

        files = {
            d.path
            for d in repository.get_pool_files(
                self.TEST_DATA / "DebianBookworm", False, set()
            )
        }

        for file in [
            "pool/main/w/wget/wget_1.21.3-1+b2_amd64.deb",
            "pool/main/w/wget/wget_1.21.3-1.debian.tar.xz",
            "pool/main/w/wget/wget_1.21.3-1.dsc",
            "pool/main/w/wget/wget_1.21.3.orig.tar.gz",
            "pool/main/w/wget/wget_1.21.3.orig.tar.gz.asc",
        ]:
            self.assertNotIn(Path(file), files)

    def test_include_source_name_mixed(self):
        config = self.get_config("DebianBookworm", config_name="mirror3.list")

        repository = self.ensure_repository(
            config.repositories["http://ftp.debian.org/debian"]
        )
        repository.mirror_path = Path("repo")

        files = {
            d.path
            for d in repository.get_pool_files(
                self.TEST_DATA / "DebianBookworm", False, set()
            )
        }

        self.assertEqual(
            {
                Path(p)
                for p in [
                    "pool/main/w/wget/wget_1.21.3-1+b2_amd64.deb",
                    "pool/main/w/wget/wget_1.21.3-1.debian.tar.xz",
                    "pool/main/w/wget/wget_1.21.3-1.dsc",
                    "pool/main/w/wget/wget_1.21.3.orig.tar.gz",
                    "pool/main/w/wget/wget_1.21.3.orig.tar.gz.asc",
                ]
            },
            files,
        )

    def test_exclude_binary_packages(self):
        config = self.get_config("DebianBookworm", config_name="mirror4.list")

        repository = self.ensure_repository(
            config.repositories["http://ftp.debian.org/debian"]
        )
        repository.mirror_path = Path("repo")

        files = {
            d.path
            for d in repository.get_pool_files(
                self.TEST_DATA / "DebianBookworm", False, set()
            )
        }

        for file in [
            "pool/main/w/wget2/libwget0_1.99.1-2.2_amd64.deb",
            "pool/main/w/wget2/wget2-dev_1.99.1-2.2_amd64.deb",
        ]:
            self.assertNotIn(Path(file), files)

    def test_include_binary_packages(self):
        config = self.get_config("DebianBookworm", config_name="mirror5.list")

        repository = self.ensure_repository(
            config.repositories["http://ftp.debian.org/debian"]
        )
        repository.mirror_path = Path("repo")

        files = {
            d.path
            for d in repository.get_pool_files(
                self.TEST_DATA / "DebianBookworm", False, set()
            )
        }

        self.assertEqual(
            {
                Path(p)
                for p in [
                    "pool/main/s/systemd/udev_252.22-1~deb12u1_amd64.deb",
                    "pool/main/w/wget2/libwget0_1.99.1-2.2_amd64.deb",
                    "pool/main/w/wget2/wget2-dev_1.99.1-2.2_amd64.deb",
                ]
            },
            files,
        )

    def test_dist_upgrader(self):
        config = self.get_config("MixedConfig", config_name="mirror1.list")

        repository = self.ensure_repository(
            config.repositories["http://ftp.debian.org/debian"]
        )
        repository.mirror_path = Path("repo")

        files = set(
            f.path
            for f in repository.get_metadata_files(
                self.TEST_DATA / "MixedConfig", False, set()
            )
        )

        base_dist_upgrader_path = Path("dists/bookworm/main/dist-upgrader-all/current")

        self.assertIn(base_dist_upgrader_path / "bookworm.tar.gz", files)
        self.assertIn(base_dist_upgrader_path / "bookworm.tar.gz.gpg", files)

        for file in Repository.DIST_UPGRADER_ANNOUNCEMENTS:
            self.assertIn(
                base_dist_upgrader_path / file,
                files,
            )
            self.assertIn(
                (base_dist_upgrader_path / file).with_suffix(".html"),
                files,
            )

    def test_dist_upgrader_updates(self):
        config = self.get_config("MixedConfig", config_name="mirror1.list")

        repository = self.ensure_repository(
            config.repositories["https://archive.ubuntu.com/ubuntu"]
        )
        repository.mirror_path = Path("repo")

        files = set(
            f.path
            for f in repository.get_metadata_files(
                self.TEST_DATA / "MixedConfig", False, set()
            )
        )

        base_dist_upgrader_path = Path(
            "dists/bionic-updates/main/dist-upgrader-all/current"
        )

        self.assertIn(base_dist_upgrader_path / "bionic.tar.gz", files)
        self.assertIn(base_dist_upgrader_path / "bionic.tar.gz.gpg", files)

        for file in Repository.DIST_UPGRADER_ANNOUNCEMENTS:
            self.assertIn(
                base_dist_upgrader_path / file,
                files,
            )
            self.assertIn(
                (base_dist_upgrader_path / file).with_suffix(".html"),
                files,
            )

    def test_packages_last_stanza(self):
        repository = self.get_repository(["main"], ["amd64"], False)

        files = repository.get_pool_files(
            self.TEST_DATA / "GitlabRegistryPackages", False, set()
        )

        single_file = (
            "pool/bullseye/main/g/gitlab-runner/gitlab-runner_14.8.1_amd64.deb"
        )
        self.assertEqual(
            {
                Path(single_file),
            },
            {d.path for d in files},
        )

    def test_sources_last_stanza(self):
        repository = self.get_repository(["main"], [], True)

        files = repository.get_pool_files(
            self.TEST_DATA / "GitlabRegistryPackages", False, set()
        )

        single_file = "pool/main/g/gitlab-runner/gitlab-runner_14.8.1.dsc"
        self.assertEqual(
            {
                Path(single_file),
            },
            {d.path for d in files},
        )

    def test_metadata_outside_root(self):
        repository = self.get_repository(["main"], [], True)

        with patch.object(
            repository,
            "_metadata_file_allowed",
            lambda *args, **kwargs: True,  # type: ignore
        ):
            files = [
                str(f.path)
                for f in repository.get_metadata_files(
                    self.TEST_DATA / "MaliciousRepository", False, set()
                )
            ]

            self.assertFalse(any(".." in path for path in files))

    def test_source_outside_root(self):
        source_parser = SourcesParser(
            self.TEST_DATA / "MaliciousRepository/repo",
            {Path("dists/test/main/source/Sources")},
            set(),
            PackageFilter(),
        )

        files = [str(f.path) for f in source_parser.parse()]

        self.assertFalse(any(".." in path for path in files))

    def test_binary_outside_root(self):
        source_parser = PackagesParser(
            self.TEST_DATA / "MaliciousRepository/repo",
            {Path("dists/test/main/binary-amd64/Packages")},
            set(),
            PackageFilter(),
        )

        files = [str(f.path) for f in source_parser.parse()]

        self.assertFalse(any(".." in path for path in files))


class TestSafePath(TestCase):
    def test_safepath(self):
        with TemporaryDirectory() as temp_folder:
            temp_folder = Path(temp_folder)

            root_folder = temp_folder / "a"
            b = Path("b")
            c = Path("../c")

            self.assertTrue(is_safe_path(root_folder, b))
            self.assertTrue(is_safe_path(root_folder, b / c))
            self.assertFalse(is_safe_path(root_folder, c))

    def test_symlink(self):
        with TemporaryDirectory() as temp_folder:
            temp_folder = Path(temp_folder)

            root_folder = temp_folder / "a"
            symlinked_folder = root_folder / "b"
            c = Path("c")
            d = Path("../d")

            root_folder.mkdir()
            symlinked_folder.symlink_to(root_folder)

            self.assertTrue(is_safe_path(root_folder, c))
            self.assertFalse(is_safe_path(root_folder, d))

            self.assertTrue(is_safe_path(symlinked_folder, c))
            self.assertFalse(is_safe_path(symlinked_folder, d))



import tempfile
from pathlib import Path
from unittest import TestCase

from apt_mirror.config import Config
from apt_mirror.repository import BaseRepository, FlatRepository, Repository


class BaseTest(TestCase):
    TEST_DATA = Path(__file__).parent / "data"
    maxDiff = None

    def get_config(self, name: str, config_name: str = "mirror.list"):
        return Config(self.TEST_DATA / name / config_name)

    def get_modified_config(
        self, name: str, extra: str, config_name: str = "mirror.list"
    ):
        with tempfile.NamedTemporaryFile("wt", encoding="utf-8") as tmp:
            with open(
                self.TEST_DATA / name / config_name, "rt", encoding="utf-8"
            ) as fp:
                tmp.write(fp.read())
                tmp.write("\n")
                tmp.write(extra)

            tmp.flush()

            return Config(Path(tmp.name))

    def ensure_repository(self, repository: BaseRepository) -> Repository:
        if not isinstance(repository, Repository):
            self.assertIsInstance(repository, Repository)
            raise RuntimeError(
                f"{repository.url} repository is not Repository instance"
            )

        return repository

    def ensure_flat_repository(self, repository: BaseRepository) -> FlatRepository:
        if not isinstance(repository, FlatRepository):
            self.assertIsInstance(repository, FlatRepository)
            raise RuntimeError(
                f"{repository.url} repository is not FlatRepository instance"
            )

        return repository



# SPDX-License-Identifer: GPL-3.0-or-later

# This file is mostly copied from the uvloop source with small changes.
# https://github.com/MagicStack/uvloop/blob/3c3bbeff3418c60b14f793a6541ad01d8036b706/uvloop/__init__.py#L117
# uvloop is dual-licensed under MIT and Apache 2.0 licenses.
# Copyright (C) 2016-present the uvloop authors and contributors.
# Copyright (C) 2024 Yuri Konotopov

import asyncio
import sys
from collections.abc import Callable, Coroutine
from typing import TYPE_CHECKING, Any, TypeVar

_T = TypeVar("_T")

try:
    import uvloop
    from uvloop import Loop

    # Copied from uvloop
    def _cancel_all_tasks(loop: asyncio.AbstractEventLoop) -> None:
        # Copied from python/cpython

        to_cancel = asyncio.all_tasks(loop)
        if not to_cancel:
            return

        for task in to_cancel:
            task.cancel()

        loop.run_until_complete(asyncio.gather(*to_cancel, return_exceptions=True))

        for task in to_cancel:
            if task.cancelled():
                continue
            if task.exception() is not None:
                loop.call_exception_handler(
                    {
                        "message": "unhandled exception during asyncio.run() shutdown",
                        "exception": task.exception(),
                        "task": task,
                    }
                )

    if TYPE_CHECKING:
        # pylint: disable=unused-argument

        def run(  # type: ignore
            main: Coroutine[Any, Any, _T],
            *,
            loop_factory: Callable[[], Loop] | None = uvloop.new_event_loop,
            debug: bool | None = None,
        ) -> _T:  # type: ignore
            """The preferred way of running a coroutine with uvloop."""

        # pylint: enable=unused-argument
    else:

        def run(
            main: Coroutine[Any, Any, Any],
            *,
            loop_factory: Callable[
                [], asyncio.AbstractEventLoop
            ] = uvloop.new_event_loop,
            **run_kwargs: str,
        ):
            async def wrapper():
                # If `loop_factory` is provided we want it to return
                # either uvloop.Loop or a subtype of it, assuming the user
                # is using `uvloop.run()` intentionally.
                loop = asyncio.get_running_loop()
                if not isinstance(loop, Loop):
                    raise TypeError("uvloop.run() uses a non-uvloop event loop")
                return await main

            vi = sys.version_info[:2]

            if vi <= (3, 10):
                loop = loop_factory()
                try:
                    asyncio.set_event_loop(loop)
                    return loop.run_until_complete(wrapper())
                finally:
                    try:
                        _cancel_all_tasks(loop)
                        loop.run_until_complete(loop.shutdown_asyncgens())
                        if hasattr(loop, "shutdown_default_executor"):
                            loop.run_until_complete(loop.shutdown_default_executor())
                    finally:
                        asyncio.set_event_loop(None)
                        loop.close()

            elif vi == (3, 11):
                with asyncio.Runner(  # pylint: disable=no-member  # type: ignore
                    loop_factory=loop_factory, **run_kwargs
                ) as runner:  # type: ignore
                    return runner.run(wrapper())  # type: ignore

            else:
                return asyncio.run(  # pylint: disable=unexpected-keyword-arg
                    wrapper(),
                    loop_factory=loop_factory,  # type: ignore
                    **run_kwargs,
                )

    UVLOOP_AVAILABLE = True

except ImportError:
    UVLOOP_AVAILABLE = False  # type: ignore

    def run(main: Coroutine[Any, Any, Any]):
        return asyncio.run(main)



# SPDX-License-Identifer: GPL-3.0-or-later

from collections.abc import AsyncIterator, Callable
from dataclasses import dataclass
from datetime import datetime


@dataclass
class DownloadResponse:
    _stream: Callable[[], AsyncIterator[bytes]] | None
    missing: bool = False
    error: str | None = None
    date: datetime | None = None
    size: int | None = None
    retry: bool | None = None

    def stream(self) -> AsyncIterator[bytes]:
        if not self._stream:
            raise RuntimeError("_stream property was not defined")

        return self._stream()



from pathlib import Path

from apt_mirror.apt_mirror import PathCleaner
from tests.base import BaseTest


class TestRepository(BaseTest):
    def test_wipe_threashold(self):
        clean_path = self.TEST_DATA / "PathClean"

        cleaner = PathCleaner(
            clean_path, set(), wipe_size_ratio=0, wipe_count_ratio=0.6
        )

        self.assertFalse(
            cleaner._clean_allowed()  # pylint: disable=W0212 # type: ignore
        )

        cleaner = PathCleaner(
            clean_path, set(), wipe_size_ratio=0.6, wipe_count_ratio=0
        )

        self.assertFalse(
            cleaner._clean_allowed()  # pylint: disable=W0212 # type: ignore
        )

        cleaner = PathCleaner(
            clean_path, {Path("1")}, wipe_size_ratio=0.7, wipe_count_ratio=0
        )

        self.assertTrue(
            cleaner._clean_allowed()  # pylint: disable=W0212 # type: ignore
        )
        cleaner = PathCleaner(
            clean_path, {Path("1")}, wipe_size_ratio=0, wipe_count_ratio=0.7
        )

        self.assertTrue(
            cleaner._clean_allowed()  # pylint: disable=W0212 # type: ignore
        )



# SPDX-License-Identifer: GPL-3.0-or-later

from collections.abc import Iterable, Sequence
from dataclasses import dataclass, field
from enum import Enum
from pathlib import Path


class HashType(Enum):
    SHA512 = "SHA512"
    SHA256 = "SHA256"
    SHA1 = "SHA1"
    MD5 = "MD5Sum"


@dataclass
class HashSum:
    type: HashType
    hash: str


class FileCompression(Enum):
    XZ = "xz"
    GZ = "gz"
    BZ2 = "bz2"
    NONE = None

    @staticmethod
    def all_compressed() -> Iterable["FileCompression"]:
        return (
            compression
            for compression in FileCompression
            if compression != FileCompression.NONE
        )

    @staticmethod
    def all_compressed_extensions() -> Iterable[str]:
        return (
            compression.file_extension
            for compression in FileCompression.all_compressed()
        )

    @property
    def file_extension(self) -> str:
        if self.value:
            return f".{self.value}"

        return ""


@dataclass
class DownloadFileCompressionVariant:
    path: Path
    compression: FileCompression
    size: int
    hashes: dict[HashType, HashSum]
    use_by_hash: bool

    def _get_hashed_path(self, hash_type: HashType):
        return (
            self.path.parent / "by-hash" / hash_type.value / self.hashes[hash_type].hash
        )

    def get_source_path(self) -> Path:
        if self.use_by_hash:
            for hash_type in HashType:
                if hash_type in self.hashes:
                    return self._get_hashed_path(hash_type)

        return self.path

    def get_all_paths(self) -> Sequence[Path]:
        paths: list[Path] = []

        if not self.use_by_hash:
            paths.append(self.path)

        if self.use_by_hash:
            paths += [self._get_hashed_path(hash_type) for hash_type in self.hashes]

        if self.use_by_hash:
            paths.append(self.path)

        return paths


@dataclass
class DownloadFile:
    path: Path
    compression_variants: dict[FileCompression, DownloadFileCompressionVariant] = field(
        default_factory=lambda: {}
    )
    check_size: bool = False
    ignore_errors: bool = False
    ignore_missing: bool = False

    @staticmethod
    def uncompressed_path(path: Path):
        if path.suffix in FileCompression.all_compressed_extensions():
            return path.with_suffix("")

        return path

    @classmethod
    def from_path(
        cls, path: Path, check_size: bool = False, ignore_missing: bool = False
    ):
        return cls(path=path, check_size=check_size, ignore_missing=ignore_missing)

    @classmethod
    def from_hashed_path(
        cls,
        path: Path,
        size: int,
        hash_type: HashType,
        hash_sum: HashSum,
        use_by_hash: bool,
    ):
        download_file = DownloadFile.from_path(cls.uncompressed_path(path))
        download_file.add_compression_variant(
            path=path,
            size=size,
            hash_type=hash_type,
            hash_sum=hash_sum,
            use_by_hash=use_by_hash,
        )

        return download_file

    def __post_init__(self):
        if not self.compression_variants:
            self.add_compression_variant(self.path, size=0)

    def add_compression_variant(
        self,
        path: Path,
        size: int,
        hash_type: HashType | None = None,
        hash_sum: HashSum | None = None,
        use_by_hash: bool = False,
    ):
        hashes: dict[HashType, HashSum] = {}
        if hash_type and hash_sum:
            hashes[hash_type] = hash_sum

        compression = FileCompression.NONE
        for _compression in FileCompression:
            if _compression.file_extension.lower() == path.suffix:
                compression = _compression
                break

        if compression == FileCompression.NONE:
            self.path = path

        compression_variant = self.compression_variants.setdefault(
            compression,
            DownloadFileCompressionVariant(
                path=path,
                compression=compression,
                size=size,
                hashes=hashes,
                use_by_hash=use_by_hash,
            ),
        )

        compression_variant.path = path
        compression_variant.size = size
        compression_variant.use_by_hash = use_by_hash
        compression_variant.hashes.update(hashes)

    def iter_variants(self):
        for compression in FileCompression:
            if compression in self.compression_variants:
                yield self.compression_variants[compression]

    @property
    def size(self):
        for compression in FileCompression:
            if compression not in self.compression_variants:
                continue

            return self.compression_variants[compression].size

        raise RuntimeError(f"{self.__class__.__name__} {self} is empty")

    def __hash__(self) -> int:
        return hash(self.path)

    def __str__(self) -> str:
        return str(self.path)

    def __repr__(self) -> str:
        return f"DownloadFile: path: {self.path}, size: {self.size}"



# SPDX-License-Identifer: GPL-3.0-or-later

import subprocess
from collections.abc import Iterator, MutableMapping
from dataclasses import dataclass, field
from pathlib import Path
from string import Template
from typing import Any

from apt_mirror.download import URL, Proxy
from apt_mirror.repository import (
    BaseRepository,
    ByHash,
    Codename,
    FlatDirectory,
    FlatRepository,
    Repository,
)

from .download.slow_rate_protector import SlowRateProtectorFactory
from .logs import LoggerFactory
from .netrc import NetRC
from .version import __version__


class RepositoryConfigException(Exception):
    pass


@dataclass
class RepositoryConfig:
    key: str
    url: URL
    arches: list[str]
    source: bool
    codenames: list[str]
    components: list[str]
    by_hash: ByHash

    @classmethod
    def from_line(cls, line: str, default_arch: str):
        log = LoggerFactory.get_logger(cls)

        repository_type, url = line.split(maxsplit=1)
        source = False

        arches: list[str] = []
        if "-" in repository_type:
            _, arch = repository_type.split("-", maxsplit=1)
            if arch != "src":
                arches.append(arch)
            else:
                source = True

        by_hash = ByHash.default()
        if url.startswith("["):
            options, url = url.split(sep="]", maxsplit=1)
            options = options.strip("[]").strip().split()
            for key, value in map(lambda x: x.split("=", maxsplit=1), options):
                match key:
                    case "arch":
                        for arch in value.split(","):
                            if arch == "src":
                                source = True
                                continue

                            if arch in arches:
                                continue

                            arches.append(arch)
                    case "by-hash":
                        try:
                            by_hash = ByHash(value)
                        except ValueError:
                            log.warning(
                                "Wrong `by-hash` value"
                                f" {value}. Affected config"
                                f" line: {line}"
                            )
                    case _:
                        continue

        url, codename = url.split(maxsplit=1)

        if not arches and not source:
            arches.append(default_arch)

        if " " in codename:
            codename, components = codename.split(maxsplit=1)
            components = components.split()
        else:
            components = []

        codenames = codename.split(",")

        if not all(c.endswith("/") for c in codenames) and not all(
            not c.endswith("/") for c in codenames
        ):
            raise RepositoryConfigException(
                f"Mixing flat and non-flat configuration for repository {url} is not"
                f" supported. Wrong codenames: {codenames}"
            )

        url = url.rstrip("/")

        return cls(
            url, URL.from_string(url), arches, source, codenames, components, by_hash
        )

    def to_repository(self) -> BaseRepository:
        if self.is_flat():
            return FlatRepository(
                url=self.url,
                clean=False,
                skip_clean=set(),
                http2_disable=False,
                mirror_dist_upgrader=False,
                mirror_path=None,
                ignore_errors=set(),
                directories=FlatRepository.FlatDirectories(
                    (
                        directory,
                        FlatDirectory(
                            self.by_hash,
                            directory,
                            self.source,
                            bool(self.arches),
                        ),
                    )
                    for directory in (Path(c.rstrip("/")) for c in self.codenames)
                ),
            )
        else:
            return Repository(
                url=self.url,
                clean=False,
                skip_clean=set(),
                http2_disable=False,
                mirror_dist_upgrader=False,
                mirror_path=None,
                ignore_errors=set(),
                codenames=Repository.Codenames(
                    (
                        codename,
                        Codename(
                            self.by_hash,
                            codename,
                            {
                                component: Codename.Component(
                                    component, self.source, self.arches
                                )
                                for component in self.components
                            },
                        ),
                    )
                    for codename in self.codenames
                ),
            )

    def update_repository(self, repository: BaseRepository):
        if isinstance(repository, FlatRepository) != self.is_flat():
            raise RepositoryConfigException(
                "Mixing of flat and non flat repositories is not supported for url"
                f" {self.url}"
            )

        for codename in self.codenames:
            if isinstance(repository, Repository):
                codename = repository.codenames.setdefault(
                    codename,
                    Codename(
                        self.by_hash,
                        codename,
                        {
                            component: Codename.Component(
                                component, self.source, self.arches
                            )
                            for component in self.components
                        },
                    ),
                )

                if codename.by_hash == ByHash.default():
                    codename.by_hash = self.by_hash

                for component in self.components:
                    component = codename.components.setdefault(
                        component,
                        Codename.Component(component, self.source, self.arches),
                    )

                    for arch in self.arches:
                        if arch not in component.arches:
                            component.arches.append(arch)

                    if self.source:
                        component.mirror_source = True

            elif isinstance(repository, FlatRepository):
                directory_path = Path(codename.rstrip("/"))
                directory = repository.directories.setdefault(
                    directory_path,
                    FlatDirectory(
                        self.by_hash, directory_path, self.source, bool(self.arches)
                    ),
                )

                if directory.by_hash == ByHash.default():
                    directory.by_hash = self.by_hash

                if not directory.mirror_binaries and bool(self.arches):
                    directory.mirror_binaries = True

                if self.source:
                    directory.mirror_source = True

    def is_flat(self):
        return any(codename.endswith("/") for codename in self.codenames)


class RepositoriesDict(MutableMapping[str, BaseRepository]):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__()
        self._dict: dict[str, BaseRepository] = {}
        self.update(dict(*args, **kwargs))

    def _find_key(self, key: Any):
        if key in self._dict:
            return key

        if isinstance(key, str):
            if key.endswith("/"):
                stripped_key = key.rstrip("/")
                if stripped_key in self._dict:
                    return stripped_key
            else:
                key_with_slash = f"{key}/"
                if key_with_slash in self._dict:
                    return key_with_slash

        return key

    def __contains__(self, key: object) -> bool:
        return self._find_key(key) in self._dict

    def __iter__(self) -> Iterator[str]:
        return iter(self._dict)

    def __len__(self) -> int:
        return len(self._dict)

    def __getitem__(self, key: str) -> Any:
        return self._dict[self._find_key(key)]

    def __setitem__(self, key: str, value: BaseRepository) -> None:
        self._dict[self._find_key(key)] = value

    def __delitem__(self, key: str) -> None:
        del self._dict[self._find_key(key)]

    def copy(self):
        return RepositoriesDict(self)


class Config:
    @dataclass
    class PackageFilter:
        include_source_name: dict[str, set[str]] = field(default_factory=dict)
        exclude_source_name: dict[str, set[str]] = field(default_factory=dict)
        include_binary_packages: dict[str, set[str]] = field(default_factory=dict)
        exclude_binary_packages: dict[str, set[str]] = field(default_factory=dict)

    DEFAULT_CONFIGFILE = "/etc/apt/mirror.list"

    def __init__(self, config_file: Path) -> None:
        self._log = LoggerFactory.get_logger(self)
        self._repositories = RepositoriesDict()

        self._files = [config_file]
        config_directory = config_file.with_name(f"{config_file.name}.d")
        if config_directory.is_dir():
            for file in config_directory.glob("*"):
                if not file.is_file() or file.suffix != ".list":
                    continue

                self._files.append(file)

        try:
            default_arch = subprocess.run(
                ["dpkg", "--print-architecture"],
                stdout=subprocess.PIPE,
                check=False,
                encoding="utf-8",
            ).stdout.strip()

            if not default_arch:
                raise FileNotFoundError()
        except FileNotFoundError:
            default_arch = "amd64"

        self._variables: dict[str, str] = {
            "defaultarch": default_arch,
            "nthreads": "20",
            "uvloop": "1",
            "base_path": "/var/spool/apt-mirror",
            "mirror_path": "$base_path/mirror",
            "skel_path": "$base_path/skel",
            "var_path": "$base_path/var",
            "etc_netrc": "/etc/apt/auth.conf",
            "cleanscript": "$var_path/clean.sh",
            "write_file_lists": "off",
            "run_postmirror": "0",
            "postmirror_script": "$var_path/postmirror.sh",
            "_contents": "1",
            "_autoclean": "0",
            "wipe_size_ratio": "0.4",
            "wipe_count_ratio": "0.4",
            "_tilde": "0",
            "limit_rate": "100m",
            "slow_rate_protection": "on",
            "slow_rate_startup": "15",
            "slow_rate": "100k",
            "unlink": "0",
            "use_proxy": "off",
            "http_proxy": "",
            "https_proxy": "",
            "proxy_user": "",
            "proxy_password": "",
            "http_user_agent": f"apt-mirror2/{__version__}",
            "no_check_certificate": "0",
            "certificate": "",
            "private_key": "",
            "ca_certificate": "",
            "prometheus_enable": "off",
            "prometheus_host": "localhost",
            "prometheus_port": "8000",
            "release_files_retries": "15",
        }

        self._parse_config_file()
        self._substitute_variables()

    def _parse_config_file(self):
        clean: list[str] = []
        skip_clean: list[str] = []
        http2_disable: list[str] = []
        mirror_dist_upgrader: list[str] = []
        mirror_paths: dict[str, Path] = {}
        ignore_errors: dict[str, set[str]] = {}
        package_filter = Config.PackageFilter()

        for file in self._files:
            with open(file, "rt", encoding="utf-8") as fp:
                for line in fp:
                    line = line.strip()

                    match line:
                        case line if line.startswith("set "):
                            _, key, value = line.split(maxsplit=2)
                            self._variables[key] = value
                        case line if line.startswith("deb"):
                            try:
                                repository_config = RepositoryConfig.from_line(
                                    line, self.default_arch
                                )
                            except ValueError:
                                self._log.warning(
                                    f"Unable to parse repository config line: {line}"
                                )
                                continue

                            repository = self._repositories.get(repository_config.key)
                            if repository:
                                repository_config.update_repository(repository)
                            else:
                                self._repositories[repository_config.key] = (
                                    repository_config.to_repository()
                                )
                        case line if line.startswith("clean "):
                            _, url = line.split()
                            clean.append(url.rstrip("/"))
                        case line if line.startswith("skip-clean "):
                            _, url = line.split()
                            skip_clean.append(url)
                        case line if line.startswith("http2-disable "):
                            _, url = line.split()
                            http2_disable.append(url)
                        case line if line.startswith("mirror_dist_upgrader "):
                            _, url = line.split()
                            mirror_dist_upgrader.append(url.rstrip("/"))
                        case line if line.startswith("mirror_path "):
                            _, url, path = line.split(maxsplit=2)
                            mirror_paths[url.rstrip("/")] = Path(path.strip("/"))
                        case line if line.startswith("ignore_errors "):
                            _, url, path = line.split(maxsplit=2)
                            ignore_errors.setdefault(url.rstrip("/"), set()).add(path)
                        case line if line.startswith("include_source_name "):
                            sources = line.split()[1:]
                            url = sources.pop(0).rstrip("/")

                            package_filter.include_source_name.setdefault(
                                url, set()
                            ).update(sources)
                        case line if line.startswith("exclude_source_name "):
                            sources = line.split()[1:]
                            url = sources.pop(0).rstrip("/")

                            package_filter.exclude_source_name.setdefault(
                                url, set()
                            ).update(sources)
                        case line if line.startswith("include_binary_packages "):
                            packages = line.split()[1:]
                            url = packages.pop(0).rstrip("/")

                            package_filter.include_binary_packages.setdefault(
                                url, set()
                            ).update(packages)
                        case line if line.startswith("exclude_binary_packages "):
                            packages = line.split()[1:]
                            url = packages.pop(0).rstrip("/")

                            package_filter.exclude_binary_packages.setdefault(
                                url, set()
                            ).update(packages)
                        case line if not line or any(
                            line.startswith(prefix) for prefix in ("#", ";")
                        ):
                            pass
                        case _:
                            self._log.warning(f"Unknown line in config: {line}")

        self._update_clean(clean)
        self._update_skip_clean(skip_clean)
        self._update_http2_disable(http2_disable)
        self._update_mirror_dist_upgrader(mirror_dist_upgrader)
        self._update_mirror_paths(mirror_paths)
        self._update_ignore_errors(ignore_errors)
        self._update_filters(package_filter)
        self._update_netrc()

    def _update_clean(self, clean: list[str]):
        for url in clean:
            if url not in self._repositories:
                self._log.warning(
                    f"Clean was specified for missing repository URL: {url}"
                )
                continue

            self._repositories[url].clean = True

    def _update_skip_clean(self, skip_clean: list[str]):
        for key in skip_clean:
            url = URL.from_string(key)
            repositories = [
                r for r in self._repositories.values() if r.url.is_part_of(url)
            ]

            for repository in repositories:
                repository.skip_clean.add(
                    Path(url.path).relative_to(Path(repository.url.path))
                )

    def _update_http2_disable(self, http2_disable: list[str]):
        for url in http2_disable:
            if url not in self._repositories:
                self._log.warning(
                    f"http2-disable is specified for missing repository URL: {url}"
                )
                continue

            self._repositories[url].http2_disable = True

    def _update_mirror_dist_upgrader(self, mirror_dist_upgrader: list[str]):
        for url in mirror_dist_upgrader:
            if url not in self._repositories:
                self._log.warning(
                    "mirror_dist_upgrader was specified for missing repository URL:"
                    f" {url}"
                )
                continue

            self._repositories[url].mirror_dist_upgrader = True

    def _update_mirror_paths(self, mirror_paths: dict[str, Path]):
        for url, path in mirror_paths.items():
            if url not in self._repositories:
                self._log.warning(
                    f"mirror_path was specified for missing repository URL: {url}"
                )
                continue

            self._repositories[url].mirror_path = path

    def _update_ignore_errors(self, ignore_errors: dict[str, set[str]]):
        for url, paths in ignore_errors.items():
            if url not in self._repositories:
                self._log.warning(
                    f"ignore_errors was specified for missing repository URL: {url}"
                )
                continue

            self._repositories[url].ignore_errors.update(paths)

    def _update_filters(
        self,
        package_filter: "Config.PackageFilter",
    ):
        for url, repository in self._repositories.items():
            repository.package_filter.include_source_name.update(
                package_filter.include_source_name.get(url, set())
            )
            repository.package_filter.exclude_source_name.update(
                package_filter.exclude_source_name.get(url, set())
            )
            repository.package_filter.include_binary_packages.update(
                package_filter.include_binary_packages.get(url, set())
            )
            repository.package_filter.exclude_binary_packages.update(
                package_filter.exclude_binary_packages.get(url, set())
            )

        for filter_name in (
            "include_source_name",
            "exclude_source_name",
            "include_binary_packages",
            "exclude_binary_packages",
        ):
            for url in getattr(package_filter, filter_name):
                if url not in self._repositories:
                    self._log.warning(
                        f"{filter_name} was specified for missing repository URL: {url}"
                    )
                    continue

    def _update_netrc(self):
        netrc = NetRC(self.etc_netrc)

        for repository in self._repositories.values():
            url = repository.url
            auth = netrc.match_machine(url)

            if auth and not url.username and not url.password:
                url.username, url.password = auth

    def _substitute_variables(self):
        max_tries = 16
        template_found = False
        while max_tries == 16 or template_found:
            template_found = False
            for key, value in self._variables.items():
                if "$" not in value:
                    continue

                self._variables[key] = Template(value).substitute(self._variables)
                template_found = True

            max_tries -= 1
            if max_tries < 1:
                raise ValueError(
                    "apt-mirror: too many substitutions while evaluating variables"
                )

    def __getitem__(self, key: str) -> str:
        if key not in self._variables:
            raise KeyError(
                f"Variable {key} is not defined in the config file {self._files[0]}"
            )

        return self._variables[key]

    def create_working_directories(self):
        for variable in ("mirror_path", "base_path", "var_path"):
            path = Path(self[variable])
            path.mkdir(parents=True, exist_ok=True)

    def init_log_files(self):
        LoggerFactory.add_log_file(None, self.var_path / "apt-mirror2.log")
        for repository in self.repositories.values():
            log_name = repository.as_filename(self.encode_tilde)
            self._add_url_log_file(repository.url, self.var_path / f"{log_name}.log")

    def _add_url_log_file(self, repository_url: URL, log_file: Path):
        LoggerFactory.add_log_file(repository_url, log_file)

    def get_bool(self, key: str) -> bool:
        return bool(self[key]) and self[key].lower() not in ("0", "off", "no")

    def get_path(self, key: str) -> Path:
        return Path(self[key])

    def get_size(self, key: str) -> int:
        suffix = self[key][-1:]

        if not suffix.isnumeric():
            value = int(self[key][:-1])
            match suffix.lower():
                case "k":
                    return value * 1024
                case "m":
                    return value * 1024 * 1024
                case _:
                    raise ValueError(
                        f"Wrong `{key}` configuration suffix: {self[key]}. Allowed"
                        " suffixes: k, m"
                    )

        return int(self[key])

    def as_environment(self) -> dict[str, str]:
        return {f"APT_MIRROR_{k.upper()}": v for k, v in self._variables.items()}

    @property
    def autoclean(self) -> bool:
        return self.get_bool("_autoclean")

    @property
    def wipe_size_ratio(self) -> float:
        return float(self._variables["wipe_size_ratio"])

    @property
    def wipe_count_ratio(self) -> float:
        return float(self._variables["wipe_count_ratio"])

    @property
    def base_path(self) -> Path:
        return self.get_path("base_path")

    @property
    def verify_ca_certificate(self) -> bool | str:
        if self.get_bool("no_check_certificate"):
            return False

        if self._variables.get("ca_certificate"):
            return self["ca_certificate"]

        return True

    @property
    def client_private_key(self) -> str:
        return self["private_key"]

    @property
    def client_certificate(self) -> str:
        return self["certificate"]

    @property
    def cleanscript(self) -> Path:
        return self.get_path("cleanscript")

    @property
    def default_arch(self):
        return self["defaultarch"]

    @property
    def encode_tilde(self):
        return self.get_bool("_tilde")

    @property
    def limit_rate(self) -> int:
        return self.get_size("limit_rate")

    @property
    def slow_rate_protector_factory(self) -> SlowRateProtectorFactory:
        return SlowRateProtectorFactory(
            self.slow_rate_protection,
            self.slow_rate_startup,
            self.slow_rate,
        )

    @property
    def slow_rate_protection(self) -> bool:
        return self.get_bool("slow_rate_protection")

    @property
    def slow_rate_startup(self) -> int:
        return int(self._variables["slow_rate_startup"])

    @property
    def slow_rate(self) -> int:
        return self.get_size("slow_rate")

    @property
    def nthreads(self) -> int:
        return int(self._variables["nthreads"])

    @property
    def mirror_path(self) -> Path:
        return self.get_path("mirror_path")

    @property
    def postmirror_script(self) -> Path:
        return self.get_path("postmirror_script")

    @property
    def repositories(self) -> RepositoriesDict:
        return self._repositories.copy()

    @property
    def write_file_lists(self):
        return self.get_bool("write_file_lists")

    @property
    def run_postmirror(self):
        return self.get_bool("run_postmirror")

    @property
    def skel_path(self) -> Path:
        return self.get_path("skel_path")

    @property
    def use_uvloop(self) -> bool:
        return self.get_bool("uvloop")

    @property
    def var_path(self) -> Path:
        return self.get_path("var_path")

    @property
    def etc_netrc(self) -> Path:
        return self.get_path("etc_netrc")

    @property
    def proxy(self) -> Proxy:
        return Proxy(
            use_proxy=self.get_bool("use_proxy"),
            http_proxy=self["http_proxy"],
            https_proxy=self["https_proxy"],
            username=self._variables.get("proxy_user"),
            password=self._variables.get("proxy_password"),
        )

    @property
    def user_agent(self) -> str:
        return self["http_user_agent"]

    @property
    def prometheus_enable(self) -> bool:
        return self.get_bool("prometheus_enable")

    @property
    def prometheus_host(self) -> str:
        return self["prometheus_host"]

    @property
    def prometheus_port(self) -> int:
        return int(self["prometheus_port"])

    @property
    def release_files_retries(self) -> int:
        return max(1, int(self["release_files_retries"]))



# SPDX-License-Identifer: GPL-3.0-or-later

from abc import ABCMeta, abstractmethod
from datetime import datetime
from typing import Any

from .format import format_size


class SlowRateException(RuntimeError):
    def __init__(self, target: str, rate: str, *args: object) -> None:
        super().__init__(f"Slow rate for `{target}`: {rate}", *args)


class BaseSlowRateProtector(metaclass=ABCMeta):
    @abstractmethod
    def rate(self, count: int): ...


class DummySlowRateProtector(BaseSlowRateProtector):
    def rate(self, count: int):
        return


class SlowRateProtector(BaseSlowRateProtector):
    def __init__(self, target: Any, rate_startup: int, slow_rate: int) -> None:
        self._startup = datetime.now()

        self._target = target
        self._rate_startup = rate_startup
        self._slow_rate = slow_rate

        self._count = 0

    def rate(self, count: int):
        passed = (datetime.now() - self._startup).seconds
        self._count += count

        if not passed or passed < self._rate_startup:
            return

        rate = self._count / passed

        if rate < self._slow_rate:
            raise SlowRateException(self._target, f"{format_size(rate)}/sec")


class SlowRateProtectorFactory:
    def __init__(self, enabled: bool, rate_startup: int, slow_rate: int) -> None:
        self._enabled = enabled
        self._rate_startup = rate_startup
        self._slow_rate = slow_rate

    def for_target(self, target: Any) -> BaseSlowRateProtector:
        if not self._enabled:
            return DummySlowRateProtector()

        return SlowRateProtector(target, self._rate_startup, self._slow_rate)



# SPDX-License-Identifer: GPL-3.0-or-later

from collections.abc import AsyncIterator
from contextlib import asynccontextmanager
from dataclasses import dataclass
from datetime import datetime, timezone
from pathlib import Path, PurePosixPath

import aioftp  # type: ignore
from aioftp.errors import StatusCodeError  # type: ignore

from ..downloader import Downloader
from ..response import DownloadResponse


class FTPFileMissingException(Exception):
    pass


@dataclass
class FTPStat:
    size: int | None = None
    date: datetime | None = None


class FTPDownloader(Downloader):
    def iter_by_block(self, ftp: aioftp.Client, path: Path):
        async def func() -> AsyncIterator[bytes]:
            async with ftp.download_stream(path) as stream:  # type: ignore
                async for data in stream.iter_by_block(  # type: ignore
                    count=self.BUFFER_SIZE
                ):
                    yield data

        return func

    @asynccontextmanager
    async def stream(self, source_path: Path):
        try:
            async with aioftp.Client.context(  # type: ignore
                self._settings.url.hostname,
                port=self._settings.url.port or aioftp.DEFAULT_PORT,
                user=self._settings.url.username or aioftp.DEFAULT_USER,
                password=self._settings.url.password or aioftp.DEFAULT_PASSWORD,
                parse_list_line_custom_first=True,
                socket_timeout=30,
            ) as ftp:
                ftp.parse_list_line_custom = self.parse_list_line_unix_links(ftp)

                await ftp.change_directory(self._settings.url.path)

                try:
                    stat = await self._get_stat(ftp, source_path)
                except FTPFileMissingException:
                    yield DownloadResponse(_stream=None, missing=True)
                    return
                except StatusCodeError as ex:
                    yield DownloadResponse(_stream=None, error=str(ex))
                    return

                yield DownloadResponse(
                    missing=False,
                    error=None,
                    date=stat.date,
                    size=stat.size,
                    _stream=self.iter_by_block(ftp, source_path),  # type: ignore
                )
        except OSError as ex:
            # https://github.com/aio-libs/aioftp/issues/173
            connection_refused = ex.errno == 111 or "111" in str(ex)
            yield DownloadResponse(
                _stream=None,
                retry=connection_refused,
                error=(str(ex) if not connection_refused else None),
            )
        except Exception as ex:  # pylint: disable=W0718
            yield DownloadResponse(
                _stream=None,
                error=f"{ex.__class__.__qualname__}: {str(ex)}",
            )

    def parse_list_line_unix_links(self, client: aioftp.Client):
        def do_parse_list_line(b: bytes) -> tuple[PurePosixPath, dict[str, str]]:
            info: dict[str, str]
            path, info = client.parse_list_line_unix(b)  # type: ignore

            s = b.decode(encoding=client.encoding).rstrip()
            if s[0] == "l":
                i = s.rindex(" -> ")
                info["link_dst"] = s[i + 4 :]

            return path, info

        return do_parse_list_line

    async def _get_stat(
        self,
        ftp: aioftp.Client,
        path: Path,  # type: ignore
    ) -> FTPStat:  # type: ignore
        ftp_stat = FTPStat()

        try:
            stat: dict[str, str] = await ftp.stat(path)  # type: ignore

            recurse_depth = 5
            link_dst: str | None = stat.get("link_dst")  # type: ignore
            while link_dst and recurse_depth > 0:
                try:
                    path: Path = path.parent / link_dst  # type: ignore

                    stat = await ftp.stat(path)  # type: ignore
                    link_dst = stat.get("link_dst")  # type: ignore
                except StatusCodeError:
                    # Give up if we can not stat link destination
                    break

                recurse_depth -= 1

            if stat.get("type") != "file":  # type: ignore
                raise FTPFileMissingException()

            try:
                if not link_dst:  # type: ignore
                    ftp_stat.size = int(stat["size"])  # type: ignore

                modify, _, _ = stat["modify"].partition(".")  # type: ignore
                if len(modify) == 14:  # type: ignore
                    ftp_stat.date = datetime.strptime(
                        modify,  # type: ignore
                        "%Y%m%d%H%M%S",
                    ).replace(tzinfo=timezone.utc)

            except (AttributeError, KeyError, ValueError):
                pass
        except (StatusCodeError, FTPFileMissingException) as ex:
            if isinstance(ex, FTPFileMissingException) or ex.received_codes[-1].matches(  # type: ignore
                "550"
            ):
                raise FTPFileMissingException() from ex
            else:
                raise

        return ftp_stat



# SPDX-License-Identifer: GPL-3.0-or-later


# Fred Cirera
# Sridhar Ratnakumar
# https://stackoverflow.com/a/1094933
def format_size(size: float, suffix: str = "B"):
    for unit in ("", "Ki", "Mi", "Gi"):
        if abs(size) < 1024.0:
            return f"{size:3.1f} {unit}{suffix}"

        size /= 1024.0

    return f"{size:.1f} Ti{suffix}"
